""" Probabilistic CKY parser for monotone, string-rewriting
Linear Context-Free Rewriting Systems. """
from __future__ import print_function
from math import exp
from collections import defaultdict, deque
import logging
import numpy as np
from discodop.agenda import EdgeAgenda, Entry
from cpython cimport PyDict_Contains
from discodop.agenda cimport Entry, EdgeAgenda
from discodop._grammar cimport Grammar
from discodop.containers cimport LCFRSEdge, ChartItem, SmallChartItem, \
		FatChartItem, new_LCFRSEdge, new_ChartItem, new_FatChartItem, Rule, \
		LexicalRule, UInt, ULong, ULLong
from discodop.bit cimport nextset, nextunset, bitcount, bitlength, \
	testbitint, anextset, anextunset, abitcount, abitlength, \
	ulongset, ulongcpy, setunion
cdef extern from "macros.h":
	int BITSLOT(int b)
	ULong BITMASK(int b)
	int BITNSLOTS(int nb)
	void SETBIT(ULong a[], int b)
	ULong TESTBIT(ULong a[], int b)
DEF SX = 1
DEF SXlrgaps = 2
DEF SLOTS = 3
cdef SmallChartItem COMPONENT = new_ChartItem(0, 0)
cdef SmallChartItem NONE = new_ChartItem(0, 0)
cdef FatChartItem FATNONE = new_FatChartItem(0)
cdef FatChartItem FATCOMPONENT = new_FatChartItem(0)


def parse(sent, Grammar grammar, tags=None, bint exhaustive=True, start=None,
		list whitelist=None, bint splitprune=False, bint markorigin=False,
		estimates=None, int beamwidth=0):
	"""
	Parse sentence, a list of tokens, optionally with gold tags, and
	produce a chart, either exhaustive or up until the viterbi parse.

	:param exhaustive: don't stop at viterbi parser, return a full chart
	:param start: integer corresponding to the start symbol that analyses
		should have, e.g., grammar.toid[b'ROOT']
	:param whitelist: a whitelist of allowed ChartItems. Anything else is not
		added to the agenda.
	:param splitprune: coarse stage used a split-PCFG where discontinuous node
		appear as multiple CFG nodes. Every discontinuous node will result
		in multiple lookups into whitelist to see whether it should be
		allowed on the agenda.
	:param markorigin: in combination with splitprune, coarse labels include an
		integer to distinguish components; e.g., CFG nodes NP*0 and NP*1
		map to the discontinuous node NP_2
	:param estimates: use context-summary estimates (heuristics, figures of
		merit) to order agenda. should be a tuple with the kind of
		estimates ('SX' or 'SXlrgaps'), and the estimates themselves in a
		4-dimensional numpy matrix. If estimates are not consistent, it is
		no longer guaranteed that the optimal parse will be found.
		experimental.
	:param beamwidth: specify the maximum number of items that will be explored
		for each particular span, on a first-come-first-served basis.
		setting to 0 disables this feature. experimental.
	"""
	cdef:
		dict beam = <dict>defaultdict(int)  # histogram of spans
		dict chart = {}  # the full chart
		list viterbi = [{} for _ in grammar.toid]  # the viterbi probabilities
		EdgeAgenda agenda = EdgeAgenda()  # the agenda
		Rule *rule
		LexicalRule lexrule
		Entry entry
		LCFRSEdge edge
		SmallChartItem item, sibling, newitem = new_ChartItem(0, 0)
		SmallChartItem goal
		double [:, :, :, :] outside
		double x = 0.0, y = 0.0, score, inside
		signed int length = 0, left = 0, right = 0, gaps = 0
		signed int lensent = len(sent), estimatetype = 0
		UInt i, lhs, blocked = 0, Epsilon = grammar.toid[b'Epsilon']
		ULong maxA = 0
		ULLong vec = 0
	if lensent >= sizeof(vec) * 8:
		return parse_longsent(sent, grammar, tags=tags, start=start,
			exhaustive=exhaustive, whitelist=whitelist, splitprune=splitprune,
			markorigin=markorigin, estimates=estimates)
	if start is None:
		start = grammar.toid[grammar.start]
	goal = new_ChartItem(start, (1ULL << len(sent)) - 1)
	if estimates is not None:
		estimatetypestr, outside = estimates
		estimatetype = {'SX': SX, 'SXlrgaps': SXlrgaps}[estimatetypestr]
	for i, word in enumerate(sent):  # add preterminals to chart
		recognized = False
		item = new_ChartItem(Epsilon, i)
		tag = tags[i].encode('ascii') if tags else None
		if estimates is not None:
			length = 1
			left = i
			gaps = 0
			right = lensent - 1 - i
		for lexrule in grammar.lexicalbyword.get(word, ()):
			# if we are given gold tags, make sure we only allow matching
			# tags - after removing addresses introduced by the DOP reduction
			if (not tags or grammar.tolabel[lexrule.lhs] == tag
					or grammar.tolabel[lexrule.lhs].startswith(tag + b'@')):
				score = lexrule.prob
				if estimatetype == SX:
					score += outside[lexrule.lhs, left, right, 0]
					if score > 300.0:
						continue
				elif estimatetype == SXlrgaps:
					score += outside[lexrule.lhs, length, left + right, gaps]
					if score > 300.0:
						continue
				newitem.label = lexrule.lhs
				newitem.vec = 1ULL << i
				tmp = process_edge(newitem, score, lexrule.prob, NULL,
						item, NONE, agenda, chart, viterbi, grammar,
						exhaustive, whitelist, False, markorigin, &blocked)
				recognized |= newitem is not tmp  # has item been blocked?
				newitem = tmp
		if not recognized and tags and tag in grammar.toid:
			lhs = grammar.toid[tag]
			score = 0.0
			if estimatetype == SX:
				score += outside[lhs, left, right, 0]
				if score > 300.0:
					continue
			elif estimatetype == SXlrgaps:
				score += outside[lhs, length, left + right, gaps]
				if score > 300.0:
					continue
			tagitem = new_ChartItem(lhs, 1ULL << i)
			agenda[tagitem] = new_LCFRSEdge(score, 0.0, NULL, item, NONE)
			chart[tagitem] = {}
			recognized = True
		elif not recognized:
			return chart, NONE, 'not covered: %r' % (tag or word, )
	while agenda.length:  # main parsing loop
		entry = agenda.popentry()
		item = <SmallChartItem>entry.key
		edge = iscore(<LCFRSEdge>entry.value)
		chart[item][edge] = edge
		(<dict>viterbi[item.label])[item] = edge
		if item.label == goal.label and item.vec == goal.vec:
			if not exhaustive:
				break
		else:
			x = edge.inside
			# unary
			if estimates is not None:
				length = bitcount(item.vec)
				left = nextset(item.vec, 0)
				gaps = bitlength(item.vec) - length - left
				right = lensent - length - left - gaps
			if beamwidth:
				if beam[item.vec] > beamwidth:
					continue
				beam[item.vec] += 1
			for i in range(grammar.numrules):
				rule = &(grammar.unary[item.label][i])
				if rule.rhs1 != item.label:
					break
				score = inside = x + rule.prob
				if estimatetype == SX:
					score += outside[rule.lhs, left, right, 0]
					if score > 300.0:
						continue
				elif estimatetype == SXlrgaps:
					score += outside[rule.lhs, length, left + right, gaps]
					if score > 300.0:
						continue
				newitem.label = rule.lhs
				newitem.vec = item.vec
				newitem = process_edge(newitem, score, inside, rule,
						item, NONE, agenda, chart, viterbi, grammar,
						exhaustive, whitelist,
						splitprune and grammar.fanout[rule.lhs] != 1,
						markorigin, &blocked)
			# binary production, child is on the right
			for i in range(grammar.numrules):
				rule = &(grammar.rbinary[item.label][i])
				if rule.rhs2 != item.label:
					break
				for I, e in (<dict>viterbi[rule.rhs1]).items():
					sibling = <SmallChartItem>I
					if beamwidth:
						if beam[sibling.vec ^ item.vec] > beamwidth:
							continue
						beam[sibling.vec ^ item.vec] += 1
					if concat(rule, sibling.vec, item.vec):
						newitem.label = rule.lhs
						newitem.vec = sibling.vec ^ item.vec
						y = (<LCFRSEdge>e).inside
						score = inside = x + y + rule.prob
						if estimatetype == SX:
							length = bitcount(item.vec)
							left = nextset(item.vec, 0)
							right = lensent - length - left
							score += outside[rule.lhs, left, right, 0]
							if score > 300.0:
								continue
						elif estimatetype == SXlrgaps:
							length = bitcount(item.vec)
							left = nextset(item.vec, 0)
							gaps = bitlength(item.vec) - length - left
							right = lensent - length - left - gaps
							score += outside[rule.lhs, length, left + right, gaps]
							if score > 300.0:
								continue
						newitem = process_edge(newitem, score, inside, rule,
								sibling, item, agenda, chart, viterbi, grammar,
								exhaustive, whitelist,
								splitprune and grammar.fanout[rule.lhs] != 1,
								markorigin, &blocked)
			# binary production, child is on the left
			for i in range(grammar.numrules):
				rule = &(grammar.lbinary[item.label][i])
				if rule.rhs1 != item.label:
					break
				for I, e in (<dict>viterbi[rule.rhs2]).items():
					sibling = <SmallChartItem>I
					if beamwidth:
						if beam[item.vec ^ sibling.vec] > beamwidth:
							continue
						beam[item.vec ^ sibling.vec] += 1
					if concat(rule, item.vec, sibling.vec):
						newitem.label = rule.lhs
						newitem.vec = item.vec ^ sibling.vec
						y = (<LCFRSEdge>e).inside
						score = inside = x + y + rule.prob
						if estimatetype == SX:
							length = bitcount(newitem.vec)
							left = nextset(newitem.vec, 0)
							right = lensent - length - left
							score += outside[rule.lhs, left, right, 0]
							if score > 300.0:
								continue
						elif estimatetype == SXlrgaps:
							length = bitcount(newitem.vec)
							left = nextset(newitem.vec, 0)
							gaps = bitlength(newitem.vec) - length - left
							right = lensent - length - left - gaps
							score += outside[rule.lhs, length, left + right, gaps]
							if score > 300.0:
								continue
						newitem = process_edge(newitem, score, inside, rule,
								item, sibling, agenda, chart, viterbi, grammar,
								exhaustive, whitelist,
								splitprune and grammar.fanout[rule.lhs] != 1,
								markorigin, &blocked)
		if agenda.length > maxA:
			maxA = agenda.length
	msg = ('agenda max %d, now %d, items %d (%d labels), edges %d, blocked %d'
		% (maxA, len(agenda), len(list(filter(None, chart.values()))),
		len(list(filter(None, viterbi))), sum(map(len, chart.values())), blocked))
	if goal not in chart:
		return chart, NONE, 'no parse ' + msg
	return chart, goal, msg


cdef inline SmallChartItem process_edge(SmallChartItem newitem, double score,
		double inside, Rule *rule, SmallChartItem left, SmallChartItem right,
		EdgeAgenda agenda, dict chart, list viterbi, Grammar grammar, bint
		exhaustive, list whitelist, bint splitprune, bint markorigin,
		UInt *blocked):
	""" Decide what to do with a newly derived edge. """
	cdef UInt a, b, cnt, label
	cdef bint inagenda
	cdef bint inchart
	cdef list componentlist
	cdef dict componentdict
	# put item in `newitem`; if item ends up in chart, newitem will be replaced
	# by a fresh object, otherwise, it can be re-used.
	inagenda = agenda.contains(newitem)
	inchart = PyDict_Contains(chart, newitem) == 1
	if not inagenda and not inchart:
		#if score > 300.0:
		#	blocked[0] += 1
		#	return newitem
		# check if we need to prune this item
		if whitelist is not None and whitelist[newitem.label] is not None:
			if splitprune:  # disc. item to be treated as several split items?
				if markorigin:
					componentlist = <list>(whitelist[newitem.label])
				else:
					componentdict = <dict>(whitelist[newitem.label])
				COMPONENT.label = 0
				b = cnt = 0
				a = nextset(newitem.vec, b)
				while a != -1:
					b = nextunset(newitem.vec, a)
					# given a=3, b=6, make bitvector: 1000000 - 1000 = 111000
					COMPONENT.vec = (1ULL << b) - (1ULL << a)
					if markorigin:
						componentdict = <dict>(componentlist[cnt])
					if PyDict_Contains(componentdict, COMPONENT) != 1:
						blocked[0] += 1
						return newitem
					a = nextset(newitem.vec, b)
					cnt += 1
			else:
				label = newitem.label
				newitem.label = 0
				if PyDict_Contains(whitelist[label], newitem) != 1:
					#or inside + <double><object>outside > 300.0):
					blocked[0] += 1
					return newitem
				newitem.label = label
		# haven't seen this item before, won't prune, add to agenda
		agenda.setitem(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
		chart[newitem] = {}
	# in agenda (maybe in chart)
	elif not exhaustive and inagenda:
		agenda.setifbetter(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
	elif (inagenda and inside < (<LCFRSEdge>(agenda.getitem(newitem))).inside):
		# item has lower score, decrease-key in agenda
		# (add old, suboptimal edge to chart if parsing exhaustively)
		edge = iscore(agenda.replace(newitem,
				new_LCFRSEdge(score, inside, rule, left, right)))
		chart[newitem][edge] = edge
	# not in agenda => must be in chart
	elif (not inagenda and inside <
				(<LCFRSEdge>(<dict>viterbi[newitem.label])[newitem]).inside):
		#re-add to agenda because we found a better score.
		#should not happen without estimates!
		agenda.setitem(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
		logging.warning('WARN: updating score in agenda: %r', newitem)
	elif exhaustive:
		# suboptimal edge
		edge = iscore(new_LCFRSEdge(score, inside, rule, left, right))
		chart[newitem][edge] = edge
	return SmallChartItem.__new__(SmallChartItem)


cdef inline bint concat(Rule *rule, ULLong lvec, ULLong rvec):
	""" Determine the compatibility of two bitvectors according to the given
	yield function. Ranges should be non-overlapping, continuous when they are
	concatenated, and adhere to the ordering in the yield function.
	The yield function indicates for each span whether it should come from the
	left or right non-terminal (0 meaning left and 1 right), and whether it is
	contiguous with the previous span.

	>>> lvec = 0b0011; rvec = 0b1000
	>>> concat(((0, ), (1, )), lvec, rvec)
	True		#discontinuous, non-overlapping, linearly ordered.
	>>> concat(((0, 1), ), lvec, rvec)
	False		#lvec and rvec are not contiguous
	>>> concat(((1, ), (0, )), lvec, rvec)
	False		#rvec's span should come after lvec's span

	The actual yield functions are encoded in a binary format;
		cf. containers.pyx
		((0, 1, 0), (1, 0)) ==> args=0b10010; lengths=0b00101
		NB: note reversal due to the way binary numbers are represented
		the least significant bit (rightmost) corresponds to the lowest
		index in the sentence / constituent (leftmost). """
	if lvec & rvec:
		return False
	cdef ULLong mask = rvec if testbitint(rule.args, 0) else lvec
	cdef size_t n
	for n in range(bitlength(rule.lengths)):
		if testbitint(rule.args, n):  # component from right vector
			if rvec & mask == 0:
				return False  # check for expected component
			rvec |= rvec - 1  # trailing 0 bits => 1 bits
			mask = rvec & (~rvec - 1)  # everything before first 0 bit => 1 bits
		else:  # component from left vector
			if lvec & mask == 0:
				return False  # check for expected component
			lvec |= lvec - 1  # trailing 0 bits => 1 bits
			mask = lvec & (~lvec - 1)  # everything before first 0 bit => 1 bits
		# zero out component
		lvec &= ~mask
		rvec &= ~mask
		if testbitint(rule.lengths, n):  # a gap
			# check that there is a gap in both vectors
			if (lvec ^ rvec) & (mask + 1):
				return False
			# increase mask to cover gap
			# get minimum of trailing zero bits of lvec & rvec
			mask = (~lvec & (lvec - 1)) & (~rvec & (rvec - 1))
		mask += 1  # e.g., 00111 => 01000
	# success if we've reached the end of both left and right vector
	return lvec == rvec == 0


def parse_longsent(sent, Grammar grammar, tags=None, start=None,
		bint exhaustive=True, list whitelist=None, bint splitprune=False, bint
		markorigin=False, estimates=None):
	"""
	Same as parse(), but accepts sentences longer than the machine word size.
	"""
	cdef:
		dict chart = {}  # the full chart
		list viterbi = [{} for _ in grammar.toid]  # the viterbi probabilities
		EdgeAgenda agenda = EdgeAgenda()  # the agenda
		Rule *rule
		LexicalRule lexrule
		Entry entry
		LCFRSEdge edge
		FatChartItem item, sibling, newitem = new_FatChartItem(0)
		FatChartItem goal
		double [:, :, :, :] outside
		double x = 0.0, y = 0.0, score, inside
		signed int length = 0, left = 0, right = 0, gaps = 0
		signed int lensent = len(sent), estimatetype = 0
		UInt i, lhs, blocked = 0, Epsilon = grammar.toid[b'Epsilon']
		ULong maxA = 0
	if start is None:
		start = grammar.toid[grammar.start]
	goal = new_FatChartItem(start)
	ulongset(goal.vec, ~0UL, BITNSLOTS(lensent) - 1)
	goal.vec[BITSLOT(lensent)] = BITMASK(lensent) - 1
	if len(sent) >= (sizeof(goal.vec) * 8):
		msg = ('sentence too long (recompile with larger value for SLOTS).'
				'input length: %d. max: %d' % (lensent, sizeof(goal.vec) * 8))
		return chart, FATNONE, msg
	if estimates is not None:
		estimatetypestr, outside = estimates
		estimatetype = {'SX': SX, 'SXlrgaps': SXlrgaps}[estimatetypestr]
	for i, word in enumerate(sent):
		recognized = False
		tag = tags[i].encode('ascii') if tags else None
		item = new_FatChartItem(Epsilon)
		item.vec[0] = i
		if estimates is not None:
			length = 1
			left = i
			gaps = 0
			right = lensent - 1 - i
		for lexrule in grammar.lexicalbyword.get(word, ()):
			# if we are given gold tags, make sure we only allow matching
			# tags - after removing addresses introduced by the DOP reduction
			if (not tags or grammar.tolabel[lexrule.lhs] == tag
				or grammar.tolabel[lexrule.lhs].startswith(tag + b'@')):
				score = lexrule.prob
				if estimatetype == SX:
					score += outside[lexrule.lhs, left, right, 0]
					if score > 300.0:
						continue
				elif estimatetype == SXlrgaps:
					score += outside[lexrule.lhs, length, left + right, gaps]
					if score > 300.0:
						continue
				newitem.label = lexrule.lhs
				ulongset(newitem.vec, 0UL, SLOTS)
				SETBIT(newitem.vec, i)
				tmp = process_fatedge(newitem, score, lexrule.prob, NULL,
						item, FATNONE, agenda, chart, viterbi, grammar,
						exhaustive, whitelist, False, markorigin, &blocked)
				recognized |= newitem is not tmp  # has item been blocked?
				newitem = tmp
		if not recognized and tags and tag in grammar.toid:
			lhs = grammar.toid[tag]
			score = 0.0
			if estimatetype == SX:
				score += outside[lhs, left, right, 0]
				if score > 300.0:
					continue
			elif estimatetype == SXlrgaps:
				score += outside[lhs, length, left + right, gaps]
				if score > 300.0:
					continue
			tagitem = new_FatChartItem(lhs)
			SETBIT(tagitem.vec, i)
			agenda[tagitem] = new_LCFRSEdge(score, 0.0, NULL, item, FATNONE)
			chart[tagitem] = {}
			recognized = True
		elif not recognized:
			msg = 'not covered: %r' % (tag or word, )
			return chart, FATNONE, msg
	while agenda.length:
		entry = agenda.popentry()
		item = <FatChartItem>entry.key
		edge = iscore(<LCFRSEdge>entry.value)
		chart[item][edge] = edge
		(<dict>viterbi[item.label])[item] = edge
		if item.label == goal.label and item.vec == goal.vec:
			if not exhaustive:
				break
		else:
			x = edge.inside
			# unary
			if estimates is not None:
				length = abitcount(item.vec, SLOTS)
				left = anextset(item.vec, 0, SLOTS)
				gaps = abitlength(item.vec, SLOTS) - length - left
				right = lensent - length - left - gaps
			for i in range(grammar.numrules):
				rule = &(grammar.unary[item.label][i])
				if rule.rhs1 != item.label:
					break
				score = inside = x + rule.prob
				if estimatetype == SX:
					score += outside[rule.lhs, left, right, 0]
					if score > 300.0:
						continue
				elif estimatetype == SXlrgaps:
					score += outside[rule.lhs, length, left + right, gaps]
					if score > 300.0:
						continue
				newitem.label = rule.lhs
				ulongcpy(newitem.vec, item.vec, SLOTS)
				newitem = process_fatedge(newitem, score, inside,
						rule, item, FATNONE, agenda, chart, viterbi,
						grammar, exhaustive, whitelist,
						splitprune and grammar.fanout[rule.lhs] != 1,
						markorigin, &blocked)
			# binary right
			for i in range(grammar.numrules):
				rule = &(grammar.rbinary[item.label][i])
				if rule.rhs2 != item.label:
					break
				for I, e in (<dict>viterbi[rule.rhs1]).items():
					sibling = <FatChartItem>I
					if fatconcat(rule, sibling.vec, item.vec):
						newitem.label = rule.lhs
						setunion(newitem.vec, sibling.vec, item.vec, SLOTS)
						y = (<LCFRSEdge>e).inside
						score = inside = x + y + rule.prob
						if estimatetype == SX:
							length = abitcount(item.vec, SLOTS)
							left = anextset(item.vec, 0, SLOTS)
							right = lensent - length - left
							score += outside[rule.lhs, left, right, 0]
							if score > 300.0:
								continue
						elif estimatetype == SXlrgaps:
							length = abitcount(item.vec, SLOTS)
							left = anextset(item.vec, 0, SLOTS)
							gaps = abitlength(item.vec, SLOTS) - length - left
							right = lensent - length - left - gaps
							score += outside[rule.lhs, length, left + right, gaps]
							if score > 300.0:
								continue
						newitem = process_fatedge(newitem, score, inside,
								rule, sibling, item, agenda, chart, viterbi,
								grammar, exhaustive, whitelist,
								splitprune and grammar.fanout[rule.lhs] != 1,
								markorigin, &blocked)
			# binary left
			for i in range(grammar.numrules):
				rule = &(grammar.lbinary[item.label][i])
				if rule.rhs1 != item.label:
					break
				for I, e in (<dict>viterbi[rule.rhs2]).items():
					sibling = <FatChartItem>I
					if fatconcat(rule, item.vec, sibling.vec):
						newitem.label = rule.lhs
						setunion(newitem.vec, item.vec, sibling.vec, SLOTS)
						y = (<LCFRSEdge>e).inside
						score = inside = x + y + rule.prob
						if estimatetype == SX:
							length = abitcount(newitem.vec, SLOTS)
							left = anextset(newitem.vec, 0, SLOTS)
							right = lensent - length - left
							score += outside[rule.lhs, left, right, 0]
							if score > 300.0:
								continue
						elif estimatetype == SXlrgaps:
							length = abitcount(newitem.vec, SLOTS)
							left = anextset(newitem.vec, 0, SLOTS)
							gaps = abitlength(newitem.vec,
									SLOTS) - length - left
							right = lensent - length - left - gaps
							score += outside[rule.lhs, length, left + right, gaps]
							if score > 300.0:
								continue
						newitem = process_fatedge(newitem, score, inside,
								rule, item, sibling, agenda, chart, viterbi,
								grammar, exhaustive, whitelist,
								splitprune and grammar.fanout[rule.lhs] != 1,
								markorigin, &blocked)
		if agenda.length > maxA:
			maxA = agenda.length
	msg = ('agenda max %d, now %d, items %d (%d labels), edges %d, blocked %d'
		% (maxA, len(agenda), len(list(filter(None, chart.values()))),
		len(list(filter(None, viterbi))), sum(map(len, chart.values())), blocked))
	if goal not in chart:
		return chart, FATNONE, 'no parse ' + msg
	return chart, goal, msg


cdef inline FatChartItem process_fatedge(FatChartItem newitem, double score,
		double inside, Rule *rule, FatChartItem left, FatChartItem right,
		EdgeAgenda agenda, dict chart, list viterbi, Grammar grammar, bint
		exhaustive, list whitelist, bint splitprune, bint markorigin,
		UInt *blocked):
	""" Decide what to do with a newly derived edge. """
	cdef UInt a, b, cnt, i
	cdef bint inagenda
	cdef bint inchart
	cdef list componentlist
	cdef dict componentdict
	# if item ends up in chart, newitem will be replaced
	# by a fresh object, otherwise, it can be re-used.
	inagenda = agenda.contains(newitem)
	inchart = PyDict_Contains(chart, newitem) == 1
	if not inagenda and not inchart:
		#if score > 300.0:
		#	blocked[0] += 1
		#	return newitem
		# check if we need to prune this item
		if whitelist is not None and whitelist[newitem.label] is not None:
			if splitprune:  # disc. item to be treated as several split items?
				if markorigin:
					componentlist = <list>(whitelist[newitem.label])
				else:
					componentdict = <dict>(whitelist[newitem.label])
				b = cnt = 0
				a = anextset(newitem.vec, b, SLOTS)
				while a != -1:
					b = anextunset(newitem.vec, a, SLOTS)
					#given a=3, b=6, make bitvector: 1000000 - 1000 = 111000
					ulongset(FATCOMPONENT.vec, 0UL, SLOTS)
					for i in range(a, b):
						SETBIT(FATCOMPONENT.vec, i)
					if markorigin:
						componentdict = <dict>(componentlist[cnt])
					if PyDict_Contains(componentdict, FATCOMPONENT) != 1:
						blocked[0] += 1
						return newitem
					a = anextset(newitem.vec, b, SLOTS)
					cnt += 1
			else:
				label = newitem.label
				newitem.label = 0
				if PyDict_Contains(whitelist[label], newitem) != 1:
					#or inside + <double><object>outside > 300.0):
					blocked[0] += 1
					return newitem
				newitem.label = label
		# haven't seen this item before, won't prune, add to agenda
		agenda.setitem(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
		chart[newitem] = {}
	# in agenda (maybe in chart)
	elif not exhaustive and inagenda:
		agenda.setifbetter(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
	elif (inagenda and inside < (<LCFRSEdge>(agenda.getitem(newitem))).inside):
		# item has lower score, decrease-key in agenda
		# (add old, suboptimal edge to chart if parsing exhaustively)
		edge = iscore(agenda.replace(newitem,
				new_LCFRSEdge(score, inside, rule, left, right)))
		chart[newitem][edge] = edge
	# not in agenda => must be in chart
	elif (not inagenda and inside <
				(<LCFRSEdge>(<dict>viterbi[newitem.label])[newitem]).inside):
		#re-add to agenda because we found a better score.
		#should not happen without estimates!
		agenda.setitem(newitem,
				new_LCFRSEdge(score, inside, rule, left, right))
		logging.warning('WARN: updating score in agenda: %r', newitem)
	elif exhaustive:
		# suboptimal edge
		edge = iscore(new_LCFRSEdge(score, inside, rule, left, right))
		chart[newitem][edge] = edge
	return FatChartItem.__new__(FatChartItem)


cdef inline bint fatconcat(Rule *rule, ULong *lvec, ULong *rvec):
	""" Determine the compatibility of two bitvectors according to the given
	yield function. """
	cdef int lpos
	cdef int rpos
	cdef UInt n
	for n in range(SLOTS):
		if lvec[n] & rvec[n]:
			return False
	lpos = anextset(lvec, 0, SLOTS)
	rpos = anextset(rvec, 0, SLOTS)
	#this algorithm was adapted from rparse, FastYFComposer.
	for n in range(bitlength(rule.lengths)):
		if testbitint(rule.args, n):
			# check if there are any bits left, and
			# if any bits on the right should have gone before
			# ones on this side
			if rpos == -1 or (lpos != -1 and lpos <= rpos):
				return False
			# jump to next gap
			rpos = anextunset(rvec, rpos, SLOTS)
			if lpos != -1 and lpos < rpos:
				return False
			# there should be a gap if and only if
			# this is the last element of this argument
			if testbitint(rule.lengths, n):
				if TESTBIT(lvec, rpos):
					return False
			elif not TESTBIT(lvec, rpos):
				return False
			# jump to next argument
			rpos = anextset(rvec, rpos, SLOTS)
		else:  # if bit == 0:
			# vice versa to the above
			if lpos == -1 or (rpos != -1 and rpos <= lpos):
				return False
			lpos = anextunset(lvec, lpos, SLOTS)
			if rpos != -1 and rpos < lpos:
				return False
			if testbitint(rule.lengths, n):
				if TESTBIT(rvec, lpos):
					return False
			elif not TESTBIT(rvec, lpos):
				return False
			lpos = anextset(lvec, lpos, SLOTS)
	# success if we've reached the end of both left and right vector
	return lpos == rpos == -1


def parse_symbolic(sent, Grammar grammar, tags=None, start=None,
		bint exhaustive=True, list whitelist=None, bint splitprune=False, bint
		markorigin=False):
	"""
	Like parse(), but disregards probabilities. """
	cdef:
		dict chart = {}  # the full chart
		list items = [deque() for _ in grammar.toid]  # items for each label
		object agenda = deque()  # the agenda
		Rule *rule
		LexicalRule lexrule
		Entry entry
		LCFRSEdge edge
		SmallChartItem item, sibling, newitem = new_ChartItem(0, 0)
		SmallChartItem goal
		signed int lensent = len(sent)
		UInt i, blocked = 0, Epsilon = grammar.toid[b'Epsilon']
		ULong maxA = 0
		ULLong vec = 0
	if lensent >= sizeof(vec) * 8:
		raise NotImplementedError('sentence too long.')
	if start is None:
		start = grammar.toid[grammar.start]
	goal = new_ChartItem(start, (1ULL << len(sent)) - 1)
	for i, word in enumerate(sent):
		recognized = False
		tag = tags[i].encode('ascii') if tags else None
		item = new_ChartItem(Epsilon, i)
		for lexrule in grammar.lexicalbyword.get(word, ()):
			# if we are given gold tags, make sure we only allow matching
			# tags - after removing addresses introduced by the DOP reduction
			if (not tags or grammar.tolabel[lexrule.lhs] == tag
				or grammar.tolabel[lexrule.lhs].startswith(tag + b'@')):
				score = lexrule.prob
				newitem.label = lexrule.lhs
				newitem.vec = 1ULL << i
				edge = new_LCFRSEdge(0.0, 0.0, NULL, item, NONE)
				agenda.append(newitem)
				chart[newitem][edge] = edge
				items[newitem.label].append(newitem)
				newitem = SmallChartItem.__new__(SmallChartItem)
		if not recognized and tags and tag in grammar.toid:
			newitem.label = grammar.toid[tag]
			newitem.vec = 1ULL << i
			edge = new_LCFRSEdge(0.0, 0.0, NULL, item, NONE)
			agenda.append(newitem)
			chart[newitem][edge] = edge
			items[newitem.label].append(newitem)
			newitem = SmallChartItem.__new__(SmallChartItem)
			recognized = True
		elif not recognized:
			return chart, NONE, 'not covered: %r' % (tag or word, )
	while agenda:
		item = agenda.pop()
		if item in chart:
			continue
		items[item.label].append(item)
		if item.label == goal.label and item.vec == goal.vec:
			if not exhaustive:
				break
		else:
			# unary
			for i in range(grammar.numrules):
				rule = &(grammar.unary[item.label][i])
				if rule.rhs1 != item.label:
					break
				newitem.label = rule.lhs
				newitem.vec = item.vec
				if newitem not in chart:
					edge = new_LCFRSEdge(0.0, 0.0, rule, item, NONE)
					agenda.append(newitem)
					chart[newitem][edge] = edge
					newitem = SmallChartItem.__new__(SmallChartItem)
			# binary right
			for i in range(grammar.numrules):
				rule = &(grammar.rbinary[item.label][i])
				if rule.rhs2 != item.label:
					break
				for I in items[rule.rhs1]:
					sibling = <SmallChartItem>I
					if concat(rule, sibling.vec, item.vec):
						newitem.label = rule.lhs
						newitem.vec = sibling.vec ^ item.vec
						if newitem not in chart:
							edge = new_LCFRSEdge(0.0, 0.0, rule, sibling, item)
							agenda.append(newitem)
							chart[newitem][edge] = edge
							newitem = SmallChartItem.__new__(SmallChartItem)
			# binary left
			for i in range(grammar.numrules):
				rule = &(grammar.lbinary[item.label][i])
				if rule.rhs1 != item.label:
					break
				for I in items[rule.rhs2]:
					sibling = <SmallChartItem>I
					if concat(rule, item.vec, sibling.vec):
						newitem.label = rule.lhs
						newitem.vec = item.vec ^ sibling.vec
						if newitem not in chart:
							edge = new_LCFRSEdge(0.0, 0.0, rule, item, sibling)
							agenda.append(newitem)
							chart[newitem][edge] = edge
							newitem = SmallChartItem.__new__(SmallChartItem)
		if agenda.length > maxA:
			maxA = agenda.length
	msg = ('agenda max %d, now %d, items %d, edges %d, blocked %d' % (
			maxA, len(agenda), len(list(filter(None, chart.values()))),
			sum(map(len, chart.values())), blocked))
	if goal not in chart:
		return chart, NONE, 'no parse ' + msg
	return chart, goal, msg


#def doinsideoutside(dict chart, ChartItem start):
#	cdef dict result = dict.fromkeys(chart)
#	getinside(result, chart, start)
#
#cdef getinside(dict result, chart, start):
#
#def newparser(sent, Grammar grammar, tags=None, start=1, bint exhaustive=True,
#		list whitelist=None, bint splitprune=False, bint markorigin=False,
#		estimates=None, int beamwidth=0):
#	# assign POS tags, unaries on POS tags
#	# parse
#	for length in range(2, len(sent) + 1):
#		for leftlength in range(1, length):
#			for left in chart[leftlength]
#				for right in chart[length - leftlength]:
#					# find rules with ... => left right
#					# can left + right concatenate?
#
#cdef inline set candidateitems(Rule rule, ChartItem item, dict chart,
#		bint left=True):
#	""" For a given rule and item, find all potential sibling items which are
#	compatible with them. Uses lookup tables with set of items for every
#	start/end point of components in items. """
#	cdef size_t n
#	cdef short pos = nextset(item.vec, 0), prevpos, x, y
#	cdef set candidates = None, temp
#	for n in range(bitlength(rule.lengths)):
#		if (testbitint(rule.args, n) == 0) == left: # the given item
#			if pos == -1:
#				return False
#			prevpos = nextunset(item.vec, pos)
#			pos = nextset(item.vec, prevpos)
#		else: # the other item for which to find candidates
#			if n and testbitint(rule.lengths, n - 1): # start gap?
#				temp = set()
#				if testbitint(rule.lengths, n): # & end gap?
#					for x in range(prevpos + 1, pos):
#						for y in range(x + 1, pos):
#							temp.update(chart.bystart[x] & chart.byend[y])
#				else:
#					for x in range(prevpos + 1, pos):
#						temp.update(chart.bystart[x] & chart.byend[pos])
#				if candidates is None:
#					candidates = set(temp)
#				else:
#					candidates &= temp
#			elif testbitint(rule.lengths, n): # end gap?
#				temp = set()
#				for x in range(prevpos + 1, pos):
#					temp.update(chart.bystart[prevpos] & chart.byend[x])
#				if candidates is None:
#					candidates = set(temp)
#				else:
#					candidates &= temp
#			else: # no gaps
#				if candidates is None:
#					candidates = chart.bystart[prevpos] & chart.byend[pos]
#				else:
#					candidates &= chart.bystart[prevpos]
#					candidates &= chart.byend[pos]
#		if not candidates:
#			break
#	return candidates if pos == -1 else set()


cdef inline LCFRSEdge iscore(LCFRSEdge e):
	""" Replace estimate with inside probability """
	e.score = e.inside
	return e


def sortfunc(a):
	if isinstance(a, SmallChartItem):
		return (bitcount((<SmallChartItem>a).vec), (<SmallChartItem>a).vec)
	elif isinstance(a, FatChartItem):
		return (abitcount((<FatChartItem>a).vec, SLOTS),
			anextset((<FatChartItem>a).vec, 0, SLOTS))
	elif isinstance(a, LCFRSEdge):
		return (<LCFRSEdge>a).inside


def pprint_chart(chart, sent, tolabel):
	""" 'Pretty print' a chart. """
	cdef ChartItem a
	cdef LCFRSEdge edge
	print('chart:')
	for a in sorted(chart, key=sortfunc):
		if chart[a] == []:
			continue
		print('%s[%s] =>' % (tolabel[a.label].decode('ascii'),
				a.binrepr(len(sent))))
		if isinstance(chart[a], float):
			continue
		for edge in sorted(chart[a], key=sortfunc):
			if edge.rule is NULL:
				print('%9.7f  %9.7f ' % (exp(-edge.inside), 1), end='')
				print("\t'%s'" % sent[edge.left.lexidx()], end='')
			else:
				print('%9.7f  %9.7f ' % (exp(-edge.inside),
						exp(-edge.rule.prob)),
						'%s[%s]' % (tolabel[edge.left.label].decode('ascii'),
						edge.left.binrepr(len(sent))), end='')
				if edge.right:
					print('\t%s[%s]' % (tolabel[edge.right.label].decode(
							'ascii'), edge.right.binrepr(len(sent))), end='')
			print()
		print()


def do(sent, grammar):
	from discodop.disambiguation import marginalize
	from operator import itemgetter
	print('sentence', sent)
	sent = sent.split()
	chart, start, _ = parse(sent, grammar)
	if len(sent) < sizeof(ULLong):
		pprint_chart(chart, sent, grammar.tolabel)
	if start:
		print('10 best parse trees:')
		mpp, _, _ = marginalize('mpp', chart, start, grammar, 10)
		for a, p in reversed(sorted(mpp.items(), key=itemgetter(1))):
			print(p, a)
		print()
		return True
	print('no parse')
	return False


def main():
	from discodop._grammar import Grammar
	cdef Rule rule
	rule.args = 0b1010
	rule.lengths = 0b1010
	assert concat(&rule, 0b100010, 0b1000100)
	assert not concat(&rule, 0b1010, 0b10100)
	rule.args = 0b101
	rule.lengths = 0b101
	assert concat(&rule, 0b10000, 0b100100)
	assert not concat(&rule, 0b1000, 0b10100)
	grammar = Grammar([((('S', 'VP2', 'VMFIN'), ((0, 1, 0), )), 1),
		((('VP2', 'VP2', 'VAINF'), ((0, ), (0, 1))), 0.5),
		((('VP2', 'PROAV', 'VVPP'), ((0, ), (1, ))), 0.5),
		((('PROAV', 'Epsilon'), ('Daruber', )), 1),
		((('VAINF', 'Epsilon'), ('werden', )), 1),
		((('VMFIN', 'Epsilon'), ('muss', )), 1),
		((('VVPP', 'Epsilon'), ('nachgedacht', )), 1)], start='S')
	print(grammar)
	assert do('Daruber muss nachgedacht werden', grammar)
	assert do('Daruber muss nachgedacht werden werden', grammar)
	assert do('Daruber muss nachgedacht werden werden werden', grammar)
	print('ungrammatical sentence:')
	assert not do('muss Daruber nachgedacht werden', grammar)  # no parse
	print('(as expected)\nlong sentence (%d words):' % 67)
	assert do('Daruber muss nachgedacht ' + ' '.join(64 * ['werden']), grammar)



r""" This file contains three main transformations:

 - A straightforward binarization: binarize(), based on NLTK code.
   Provides some additional Markovization options.
 - An optimal binarization for LCFRS: optimalbinarize()
   Cf. Gildea (2010): Optimal parsing strategies for linear
   context-free rewriting systems.
 - Converting discontinuous trees to continuous trees and back:
   splitdiscnodes(). Cf. Boyd (2007): Discontinuity revisited. """
# Original notice:
# Natural Language Toolkit: Tree Transformations
#
# Copyright (C) 2005-2007 Oregon Graduate Institute
# Author: Nathan Bodenstab <bodenstab@cslu.ogi.edu>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT
from __future__ import print_function
import re
import sys
from itertools import count
from collections import defaultdict, Set, Iterable
if sys.version[0] >= '3':
	basestring = str  # pylint: disable=W0622,C0103
from discodop.tree import Tree, ImmutableTree
from discodop.grammar import ranges
try:
	from discodop.bit import fanout as bitfanout
except ImportError:
	def bitfanout(arg):
		""" Slower version. """
		prev, result = arg, 0
		while arg:
			arg &= arg - 1
			if ((prev - arg) << 1) & prev == 0:
				result += 1
			prev = arg
		return result

USAGE = """Treebank binarization and conversion
Usage: %s [options] <action> [input [output]]
where input and output are treebanks; standard in/output is used if not given.
action is one of:
    none
    binarize [-h x] [-v x] [--factor=left|right]
    optimalbinarize [-h x] [-v x]
    unbinarize
    introducepreterminals
    splitdisc [--markorigin]
    mergedisc


options may consist of (* marks default option):
  --inputfmt|--outputfmt=[*export|discbracket|bracket|alpino]
  --inputenc|--outputenc=[*UTF-8|ISO-8859-1|...]
  --slice=n:m    select a range of sentences from input starting with n,
                 up to but not including m; as in Python, n or m can be left
                 out or negative, and the first index is 0.
  --factor=[left|*right]  whether binarization factors to the left or right
  -h n           horizontal markovization. default: infinite (all siblings)
  -v n           vertical markovization. default: 1
  --headrules=x  turn on head finding; affects binarization.
                 reads rules from file "x" (e.g., "negra.headrules").
  --markheads    mark heads with '^' in phrasal labels.
  --punct=x      possible options:
                 remove: remove any punctuation.
                 move: re-attach punctuation to nearest constituent to minimize
                       discontinuity.
                 restore: attach punctuation under root node.
  --functions=x  'leave': (default): leave syntactic labels as is,
                 'remove': strip away hyphen-separated function labels
                 'add': concatenate syntactic categories with functions,
                 'replace': replace syntactic labels w/grammatical functions.
  --morphology=x 'no' (default): use POS tags as preterminals
                 'add': concatenate morphological information to POS tags,
                     e.g., DET/sg.def
                 'replace': use morphological information as preterminal label
                 'between': insert node with morphological information between
                     POS tag and word, e.g., (DET (sg.def the))
  --lemmas       insert node with lemma between word and POS tag.

Note: The formats 'conll' and 'mst' do an unlabeled dependency conversion and
    require all constituents to have a child with HD as one of its function
    labels, or the use of heuristic head rules. """ % sys.argv[0]


def binarize(tree, factor='right', horzmarkov=999, vertmarkov=1,
		childchar='|', parentchar='^', headmarked=None, headidx=None,
		tailmarker='', leftmostunary=False, rightmostunary=False, threshold=2,
		pospa=False, artpa=True, reverse=False, ids=None):
	""" Binarize an NLTK Tree object.

	:param factor: "left" or "right". Determines whether binarization proceeds
			from left to right or vice versa.
	:param horzmarkov: amount of horizontal context in labels. Default is
			infinity, such that now new generalization are introduced by the
			binarization.
	:param vertmarkov: number of ancestors to include in labels.
			NB: 1 means only the direct parent, as in a normal tree.
	:param headidx: if specified, the label of the head node is always included
			as an additional horizontal sibling; use 0 or -1 for first or last
			node respectively.
	:param headmarked: when given a string, the occurrence of this string in a
			label signifies that thte node is the head; the direction of
			binarization will be switched when it is encountered, to enable a
			head-outward binarization. NB: for discontinuous trees this is not
			necessary, as the order of children can be freely adjusted to
			achieve the same effect.
	:param leftmostunary, rightmostunary: introduce a unary production for the
			first/last child. When h=1, this enables the same generalizations
			for the first & last non-terminals as for other siblings.
	:param tailmarker: when given a non-empty string, add this to artificial
			nodes introducing the last symbol. This is useful when the last
			symbol is the head node, ensuring that it is not exchangeable with
			other non-terminals.
	:param reverse: reverse direction of the horizontal markovization;
			e.g.: (A (B ) (C ) (D )) ...becomes:
			left:  (A (A|<D> (A|<C-D> (A|<B-C> (B )) (C )) (D )))
			right: (A (A|<B> (B ) (A|<B-C> (C ) (A|<C-D> (D )))))
			in this way the markovization represents the history of the
			nonterminals that have *already* been parsed, instead of those
			still to come (assuming bottom-up parsing).
	:param pospa: whether to add parent annotation to POS nodes.
	:param artpa: whether to add parent annotation to the artificial nodes
			introduced by the binarization.
	:param ids: a function to provide artificial node labels, instead of
			combining labels of sibling nodes. Disables Markovization.
	:param threshold: constituents with more than this number of children are
			factored; i.e., for a value of 2, do a normal binarization; for a
			value of 1, also factor binary productions to include an artificial
			node, etc.

	>>> sent = 'das muss man jetzt machen'.split()
	>>> treestr = '(S (VP (PDS 0) (ADV 3) (VVINF 4)) (PIS 2) (VMFIN 1))'
	>>> origtree = Tree(treestr); tree = Tree(treestr)
	>>> print(binarize(tree, horzmarkov=0, tailmarker=''))
	(S (VP (PDS 0) (VP|<> (ADV 3) (VVINF 4))) (S|<> (PIS 2) (VMFIN 1)))
	>>> tree = unbinarize(tree); assert tree == origtree

	>>> print(binarize(tree, horzmarkov=1, tailmarker=''))
	(S (VP (PDS 0) (VP|<ADV> (ADV 3) (VVINF 4))) (S|<PIS> (PIS 2) (VMFIN 1)))

	>>> tree = unbinarize(tree); assert tree == origtree
	>>> print(binarize(tree, horzmarkov=1, leftmostunary=False,
	... rightmostunary=True, tailmarker=''))
	(S (VP (PDS 0) (VP|<ADV> (ADV 3) (VP|<VVINF> (VVINF 4)))) (S|<PIS>
		(PIS 2) (S|<VMFIN> (VMFIN 1))))

	>>> tree = unbinarize(tree); assert tree == origtree
	>>> print(binarize(tree, horzmarkov=1, leftmostunary=True,
	... rightmostunary=False, tailmarker=''))
	(S (S|<VP> (VP (VP|<PDS> (PDS 0) (VP|<ADV> (ADV 3) (VVINF 4)))) (S|<PIS>
		(PIS 2) (VMFIN 1))))

	>>> tree = unbinarize(tree); assert tree == origtree
	>>> print(binarize(tree, factor='left', horzmarkov=2, tailmarker=''))
	(S (S|<PIS,VMFIN> (VP (VP|<ADV,VVINF> (PDS 0) (ADV 3)) (VVINF 4))
		(PIS 2)) (VMFIN 1))

	>>> tree = unbinarize(tree); assert tree == origtree
	>>> print(binarize(tree, horzmarkov=1, vertmarkov=3, leftmostunary=True,
	... rightmostunary=False, tailmarker='', pospa=True))
	(S (S|<VP> (VP^<S> (VP|<PDS>^<S> (PDS^<VP,S> 0) (VP|<ADV>^<S> (ADV^<VP,S>
		3) (VVINF^<VP,S> 4)))) (S|<PIS> (PIS^<S> 2) (VMFIN^<S> 1))))

	>>> tree = Tree('(S (A 0) (B 1) (C 2) (D 3) (E 4) (F 5))')
	>>> print(binarize(tree, tailmarker='', reverse=False))
	(S (A 0) (S|<B,C,D,E,F> (B 1) (S|<C,D,E,F> (C 2) (S|<D,E,F> (D 3)
		(S|<E,F> (E 4) (F 5)))))) """
	# assume all nodes have homogeneous children, terminals have no siblings
	assert factor in ('left', 'right')
	treeclass = tree.__class__
	leftmostunary = 1 if leftmostunary else 0

	# Traverse tree depth-first keeping a list of ancestor nodes to the root.
	agenda = [(tree, [tree.label])]
	while agenda:
		node, parent = agenda.pop()
		if not isinstance(node, Tree):
			continue
		# parent annotation
		parentstring = ''
		originallabel = node.label if vertmarkov else ''
		if vertmarkov > 1 and node is not tree and (
				pospa or isinstance(node[0], Tree)):
			parentstring = '%s<%s>' % (parentchar, ','.join(parent))
			node.label += parentstring
			parent = [originallabel] + parent[:vertmarkov - 2]
			if not artpa:
				parentstring = ''

		# add children to the agenda before we mess with them
		agenda.extend((child, parent) for child in node)

		# binary form factorization
		if len(node) <= threshold:
			continue
		elif 1 <= len(node) <= 2:
			if not isinstance(node[0], Tree):
				continue
			# insert an initial artificial nonterminal
			if ids is None:
				siblings = '' if headidx is None else node[headidx].label + ';'
				siblings += ','.join(child.label for child in node[:horzmarkov])
			else:
				siblings = str(next(ids))
			newnode = treeclass('%s%s<%s>%s' % (originallabel, childchar,
					siblings, parentstring), node)
			node[:] = [newnode]
		else:
			if isinstance(node[0], Tree):
				childlabels = [child.label for child in node]
			else:
				childlabels = []
			childnodes = list(node)
			numchildren = len(childnodes)
			assert not headmarked or headidx is None
			headmarkedidx = 0

			# insert an initial artificial nonterminal
			if factor == 'right':
				start = 0
				end = min(1, horzmarkov) if reverse else horzmarkov
			else:  # factor == 'left'
				start = ((numchildren - min(1, horzmarkov))
						if reverse else horzmarkov)
				end = numchildren
			if ids is None:
				siblings = '' if headidx is None else childlabels[headidx] + ';'
				siblings += ','.join(childlabels[start:end])
			else:
				siblings = str(next(ids))
			newnode = treeclass('%s%s<%s>%s' % (originallabel, childchar,
											siblings, parentstring), [])
			node[:] = []
			if leftmostunary:
				node.append(newnode)
				node = newnode
			curnode = node

			for i in range(1, numchildren - (not rightmostunary)):
				marktail = tailmarker if i + 1 == numchildren else ''
				newnode = treeclass('', [])
				if factor == 'right':
					if reverse:
						start = max(i - horzmarkov + 1, 0)
						end = i + 1
					else:
						start = i
						end = i + horzmarkov
					curnode[:] = [childnodes.pop(0), newnode]
				else:  # factor == 'left':
					start = headmarkedidx + numchildren - i - 1
					end = start + horzmarkov
					curnode[:] = [newnode, childnodes.pop()]
				# switch direction upon encountering the head
				if headmarked and headmarked in childlabels[i]:
					headmarkedidx = i
					factor = 'right' if factor == 'left' else 'left'
					start = headmarkedidx + numchildren - i - 1
					end = start + horzmarkov
				if ids is None:
					siblings = '' if headidx is None else childlabels[headidx] + ';'
					siblings += ','.join(childlabels[start:end])
				else:
					siblings = str(next(ids))
				newnode.label = '%s%s<%s>%s%s' % (originallabel, childchar,
						siblings, parentstring, marktail)
				curnode = newnode
			assert len(childnodes) == 1 + (not rightmostunary)
			curnode.extend(childnodes)
	return tree


def unbinarize(tree, expandunary=True, childchar='|', parentchar='^',
		unarychar='+'):
	""" Restore a binarized tree to the original n-ary tree.
	Modifies tree in-place.
	NB: a malformed node such as (X|<Y> ) which is not supposed to be empty
	will be silently discarded. """
	# increase robustness
	childchar += '<'
	parentchar += '<'
	treeclass = tree.__class__
	# Traverse the tree-depth first keeping a pointer to the parent for
	# modification purposes.
	agenda = [(tree, [])]
	while agenda:
		node, parent = agenda.pop()
		if isinstance(node, Tree):
			# if the node contains the 'childchar' character it means that it
			# is an artificial node and can be removed, although we still
			# need to move its children to its parent
			childindex = node.label.find(childchar)
			if childindex != -1:
				# go by identity instead of equality
				n = [id(a) for a in parent].index(id(node))
				parent[n:n + 1] = node
			else:
				parentindex = node.label.find(parentchar)
				if parentindex != -1:
					# strip the node name of the parent annotation
					node.label = node.label[:parentindex]
				# expand collapsed unary productions
				if expandunary:
					unaryindex = node.label.find(unarychar)
					if unaryindex != -1:
						newnode = treeclass(
								node.label[unaryindex + 1:], node[:])
						node.label = node.label[:unaryindex]
						node[:] = [newnode]
				parent = node  # non-binarized node, move on to next parent
			agenda.extend((child, parent) for child in node)
	return tree


def collapse_unary(tree, collapsepos=False, collapseroot=False, joinchar='+'):
	""" Collapse subtrees with a single child (i.e., unary productions) into a
	new non-terminal (Tree node) joined by 'joinchar'. This is useful when
	working with algorithms that do not allow unary productions, and completely
	removing the unary productions would result in loss of useful information.
	The tree is modified in-place.

	:param collapsepos: when False (default), do not collapse preterminals
	:param collapseroot: when False (default) do not modify the root production
		if it is unary; e.g., TOP -> productions for the Penn WSJ treebank.
	:param joinchar: A string used to connect collapsed node values """
	agenda = [tree]
	if not collapseroot and isinstance(tree, Tree) and len(tree) == 1:
		agenda = [tree[0]]

	# depth-first traversal of tree
	while agenda:
		node = agenda.pop()
		if isinstance(node, Tree):
			if (len(node) == 1 and isinstance(node[0], Tree)
					and (collapsepos or isinstance(node[0, 0], Tree))):
				node.label += joinchar + node[0].label
				node[0:] = [child for child in node[0]]
				# since we assigned the child's children to the current node,
				# evaluate the current node again
				agenda.append(node)
			else:
				agenda.extend(node)
	return tree


def introducepreterminals(tree, ids=None):
	""" Introduce preterminals with artificial POS-tags where needed
	(i.e., for every terminal with siblings.)

	>>> tree = Tree('(S (X a b (CD c d) e))')
	>>> print(introducepreterminals(tree))
	(S (X (X/a a) (X/b b) (CD (CD/c c) (CD/d d)) (X/e e))) """
	assert isinstance(tree, Tree)
	treeclass = tree.__class__
	agenda = [tree]
	while agenda:
		node = agenda.pop()
		hassiblings = len(node) > 1
		for n, child in enumerate(node):
			if isinstance(child, Tree):
				agenda.append(child)
			elif hassiblings:
				node[n] = treeclass('%s/%s' % (
					node.label if ids is None else next(ids), child), [child])
	return tree


def getbits(bitset):
	""" Iterate over the indices of set bits in a bitset. """
	n = 0
	while bitset:
		if bitset & 1:
			yield n
		elif not bitset:
			break
		bitset >>= 1
		n += 1


def fanout(tree):
	""" Get the fan-out of a constituent. Requires the presence of a bitset
	attribute. """
	return bitfanout(tree.bitset) if isinstance(tree, Tree) else 1


def complexity(tree):
	""" The degree of the time complextiy of parsing with this rule.
	Cf. Gildea (2011). """
	return fanout(tree) + sum(map(fanout, tree))


def complexityfanout(tree):
	""" Combination of complexity and fan-out, where the latter is used
	to break ties in the former. """
	return (fanout(tree) + sum(map(fanout, tree)), fanout(tree))


def fanoutcomplexity(tree):
	""" Combination of fan-out and complexity, where the latter is used
	to break ties in the former. """
	return (fanout(tree), fanout(tree) + sum(map(fanout, tree)))


def addbitsets(tree):
	""" Turn tree into an ImmutableTree and add a bitset attribute to each
	constituent to avoid slow calls to leaves(). """
	if isinstance(tree, basestring):
		result = ImmutableTree.parse(tree, parse_leaf=int)
	elif isinstance(tree, ImmutableTree):
		result = tree
	elif isinstance(tree, Tree):
		result = tree.freeze()
	else:
		raise ValueError('expected string or tree object')
	for a in result.subtrees():
		a.bitset = sum(1 << n for n in a.leaves())
	return result


def getyf(left, right):
	""" Given two trees with bitsets, return a string representation of
	their yield function, e.g., ';01,10'. """
	result = [';']
	cur = ','
	for n in range(max(left.bitset.bit_length(), right.bitset.bit_length())):
		mask = 1 << n
		if left.bitset & mask:
			if cur != '0':
				cur = '0'
				result.append(cur)
		elif right.bitset & mask:
			if cur != '1':
				cur = '1'
				result.append(cur)
		elif cur != ',':
			cur = ','
			result.append(cur)
	return ''.join(result)


def factorconstituent(node, sep='|', h=999, factor='right',
		markfanout=False, markyf=False, ids=None, threshold=2):
	""" Binarize one constituent with a left/right factored binarization.
	Children remain unmodified. Bottom-up version. Nodes must be immutable
	and contain bitsets; use addbitsets().
	By default construct artificial labels using labels of child nodes.
	When markyf is True, each artificial label will include the yield function;
	this is necessary for a 'normal form' binarization that is equivalent to the
	original. When ids is given, it is used both as an interator (for new unique
	labels) and as a dictionary (to re-use labels). The first ID in a
	binarization will always be unique, while the others will be re-used for
	the same combination of labels and yield function. """
	if len(node) <= threshold:
		return node
	elif 1 <= len(node) <= 2:
		if ids is None:
			key = '%s%s' % (','.join(child.label for child in node[:h]),
					getyf(*node) if markyf else '')
		else:
			key = str(next(ids))
		newlabel = '%s%s<%s>' % (node.label, sep, key)

		result = ImmutableTree(node.label, [ImmutableTree(newlabel, node)])
		result.bitset = node.bitset
	else:
		if factor == 'right':
			prev = node[-1]
			rng = range(len(node) - 2, 0, -1)
		elif factor == 'left':
			prev = node[0]
			rng = range(1, len(node) - 1)
		else:
			raise ValueError
		for i in rng:
			newbitset = node[i].bitset | prev.bitset
			if factor == 'right' and (ids is None or i > 1):
				key = ','.join(child.label for child in node[i:i + h])
				if markyf:
					key += getyf(node[i], prev)
				if ids is not None:
					key = str(ids[key])
			elif factor == 'left' and (ids is None or i < len(node) - 2):
				key = ','.join(child.label
						for child in node[max(0, i - h + 1):i + 1])
				if markyf:
					key += getyf(prev, node[i])
				if ids is not None:
					key = str(ids[key])
			else:
				key = str(next(ids))
			newlabel = '%s%s<%s>' % (node.label, sep, key)
			if markfanout:
				nodefanout = bitfanout(newbitset)
				if nodefanout > 1:
					newlabel += '_' + str(nodefanout)
			prev = ImmutableTree(newlabel,
					[node[i], prev] if factor == 'right' else [prev, node[i]])
			prev.bitset = newbitset
		result = ImmutableTree(node.label,
				[node[0], prev] if factor == 'right' else [prev, node[-1]])
		result.bitset = (node[0].bitset if factor == 'right'
				else node[-1].bitset) | prev.bitset
	return result


def minimalbinarization(tree, score, sep='|', head=None, parentstr='', h=999):
	""" Implementation of Gildea (2010): Optimal parsing strategies for
	linear context-free rewriting systems.  Expects an immutable tree where
	the terminals are integers corresponding to indices, with a special
	bitset attribute to avoid having to call leaves() repeatedly.
	The bitset attribute can be added with addbitsets()

	:param tree: the tree for which the optimal binarization of its top
		production will be searched.
	:param score: a function from binarized trees to some value, where lower is
		better (the value can be numeric or anything else which supports
		comparisons).
	:param head: an optional index of the head node, specifying it enables
		head-driven binarization (which constrains the possible binarizations).

	>>> tree = '(X (A 0) (B 1) (C 2) (D 3) (E 4))'
	>>> tree1 = addbitsets(tree)
	>>> tree2 = binarize(Tree.parse(tree, parse_leaf=int))
	>>> minimalbinarization(tree1, complexityfanout, head=2) == tree2
	True
	>>> tree = addbitsets('(A (B1 (t 6) (t 13)) (B2 (t 3) (t 7) (t 10)) '
	... '(B3 (t 1) (t 9) (t 11) (t 14) (t 16)) (B4 (t 0) (t 5) (t 8)))')
	>>> a = minimalbinarization(tree, complexityfanout)
	>>> b = minimalbinarization(tree, fanoutcomplexity)
	>>> print(max(map(complexityfanout, a.subtrees())))
	(14, 6)
	>>> print(max(map(complexityfanout, b.subtrees())))
	(15, 5) """
	def newproduction(a, b):
		""" return a new 'production' (here a tree) combining a and b """
		if head is not None:
			siblings = (nonterms[a] | nonterms[b])[:h]
		else:
			siblings = getbits(nonterms[a] | nonterms[b])
		newlabel = '%s%s<%s>%s' % (tree.label, sep,
				','.join(labels[x] for x in siblings), parentstr)
		new = ImmutableTree(newlabel, [a, b])
		new.bitset = a.bitset | b.bitset
		return new
	if len(tree) <= 2:
		return tree
	#don't bother with optimality if this particular node is not discontinuous
	#do default right factored binarization instead
	elif fanout(tree) == 1 and all(fanout(a) == 1 for a in tree):
		return factorconstituent(tree, sep=sep, h=h)
	from discodop.agenda import Agenda
	labels = [a.label for a in tree]
	#the four main datastructures:
	#the agenda is a priority queue of partial binarizations to explore
	#the first complete binarization that is dequeued is the optimal one
	agenda = Agenda()
	#the working set contains all the optimal partial binarizations
	#keys are binarizations, values are their scores
	workingset = {}
	#for each of the optimal partial binarizations, this dictionary has
	#a bitset that describes which non-terminals from the input it covers
	nonterms = {}
	# reverse lookup table for nonterms (from bitsets to binarizations)
	revnonterms = {}
	#the goal is a bitset that covers all non-terminals of the input
	goal = (1 << len(tree)) - 1
	if head is None:
		for n, a in enumerate(tree):
			nonterms[a] = 1 << n
			revnonterms[nonterms[a]] = a
			workingset[a] = score(a) + (0,)
			agenda[a] = workingset[a]
	else:
		# head driven binarization:
		# add all non-head nodes to the working set,
		# add all combinations of non-head nodes with head to agenda
		# caveat: Crescenzi et al. (2011) show that this problem is NP hard.
		hd = tree[head]
		goal = OrderedSet(range(len(tree)))
		for n, a in enumerate(tree):
			nonterms[a] = OrderedSet([n])
			revnonterms[nonterms[a]] = a
			if n != head:
				workingset[a] = score(a) + (0,)
		for n, a in enumerate(tree):
			if n == head:
				continue
			# (add initial unary here)
			p = newproduction(a, hd)
			x = score(p)
			agenda[p] = workingset[p] = x + (x[0],)
			nonterms[p] = nonterms[a] | nonterms[hd]
			revnonterms[nonterms[p]] = p
	while agenda:
		p, x = agenda.popitem()
		if nonterms[p] == goal:
			# (add final unary here)
			p = ImmutableTree(tree.label, p[:])
			p.bitset = tree.bitset
			return p
		for p1, y in list(workingset.items()):
			if p1 not in workingset:
				continue
			# this is inefficient. we should have a single query for all
			# items not overlapping with p
			elif nonterms[p] & nonterms[p1]:
				continue
			# if we do head-driven binarization, add one nonterminal at a time
			if head is None:
				p2 = newproduction(p, p1)
				p2nonterms = nonterms[p] | nonterms[p1]
			elif len(nonterms[p1]) == 1:
				p2 = newproduction(p1, p)
				p2nonterms = nonterms[p1] | nonterms[p]
			elif len(nonterms[p]) == 1:
				p2 = newproduction(p, p1)
				p2nonterms = nonterms[p] | nonterms[p1]
			else:
				continue
			scorep2 = score(p2)
			# important: the score is the maximum score up till now
			x2 = max((scorep2, y[:-1], x[:-1]))
			# add the sum of all previous parsing complexities as last item
			x2 += (scorep2[0] + x[-1] + y[-1],)
			#if new or better:
			# should we allow item when score is equal?
			if (p2nonterms not in revnonterms
				or workingset[revnonterms[p2nonterms]] > x2):
				if p2nonterms in revnonterms:
					a = revnonterms[p2nonterms]
					del nonterms[a], workingset[a]
					if a in agenda:
						del agenda[a]
				nonterms[p2] = p2nonterms
				revnonterms[p2nonterms] = p2
				agenda[p2] = workingset[p2] = x2
	raise ValueError


def optimalbinarize(tree, sep='|', headdriven=False, h=None, v=1):
	""" Recursively binarize a tree optimizing for complexity.
	v=0 is not implemented.
	Setting h to a nonzero integer restricts the possible binarizations
	to head driven binarizations. """
	if h is None:
		tree = Tree.convert(tree)
		for a in list(tree.subtrees(lambda x: len(x) > 1))[::-1]:
			a.sort(key=lambda x: x.leaves())
	return recbinarizetree(addbitsets(tree), sep, headdriven, h or 999, v, ())


def recbinarizetree(tree, sep, headdriven, h, v, ancestors):
	""" postorder / bottom-up binarization """
	if not isinstance(tree, Tree):
		return tree
	parentstr = '^<%s>' % (','.join(ancestors[:v - 1])) if v > 1 else ''
	newtree = ImmutableTree(tree.label + parentstr,
		[recbinarizetree(t, sep, headdriven, h, v, (tree.label,) + ancestors)
			for t in tree])
	newtree.bitset = tree.bitset
	return minimalbinarization(newtree, complexityfanout, sep,
		parentstr=parentstr, h=h, head=(len(tree) - 1) if headdriven else None)


def disc(node):
	""" Test whether a particular node is locally discontinuous, i.e., whether
	its yield consists of two or more non-adjacent strings. Nodes can be
	continuous even if some of their children are discontinuous. """
	if not isinstance(node, Tree):
		return False
	return len(list(ranges(sorted(node.leaves())))) > 1


def addfanoutmarkers(tree):
	""" Modifies tree so that the label of each node with a fanout > 1 contains
	a marker "_n" indicating its fanout. """
	for st in tree.subtrees():
		leaves = set(st.leaves())
		thisfanout = len([a for a in sorted(leaves) if a - 1 not in leaves])
		if thisfanout > 1 and not st.label.endswith('_%d' % thisfanout):
			st.label += '_%d' % thisfanout
	return tree


def removefanoutmarkers(tree):
	""" Remove fanout marks. """
	for a in tree.subtrees(lambda x: '_' in x.label):
		a.label = a.label.rsplit('_', 1)[0]
	return tree


def postorder(tree, f=lambda _: True):
	""" Do a postorder traversal of tree; similar to Tree.subtrees(),
	but Tree.subtrees() does a preorder traversal. """
	for child in tree:
		if isinstance(child, Tree):
			for a in filter(f, postorder(child)):
				yield a
	if f(tree):
		yield tree


def canonicalize(tree):
	""" restore canonical linear precedence order. tree modified in-place. """
	for a in postorder(tree, lambda n: len(n) > 1):
		a.sort(key=lambda n: n.leaves())
	return tree


def contsets(nodes):
	""" partition children into continuous subsets

	>>> tree = Tree.parse(
	... "(VP (PP (APPR 0) (ART 1) (NN 2)) (CARD 4) (VVPP 5))", parse_leaf=int)
	>>> print(list(contsets(tree)))
	[[Tree('PP', [Tree('APPR', [0]), Tree('ART', [1]), Tree('NN', [2])])],
	[Tree('CARD', [4]), Tree('VVPP', [5])]] """
	rng, subset = -1, []
	mins = {min(a.leaves()) if isinstance(a, Tree) else a: a for a in nodes}
	leaves = [a for child in nodes for a in child.leaves()]

	for a in sorted(leaves):
		if rng >= 0 and a != rng + 1:
			yield subset
			subset = []
		if a in mins:
			subset.append(mins[a])
		rng = a
	if subset:
		yield subset


def splitdiscnodes(tree, markorigin=False):
	""" Boyd (2007): Discontinuity revisited.
	markorigin=False: VP* (bare label)
	markorigin=True: VP*1 (add index)

	>>> tree = Tree.parse('(S (VP (VP (PP (APPR 0) (ART 1) (NN 2)) (CARD 4)'
	... '(VVPP 5)) (VAINF 6)) (VMFIN 3))', parse_leaf=int)
	>>> print(splitdiscnodes(tree.copy(True)))
	(S (VP* (VP* (PP (APPR 0) (ART 1) (NN 2)))) (VMFIN 3) (VP* (VP* (CARD 4)
		(VVPP 5)) (VAINF 6)))
	>>> print(splitdiscnodes(tree, markorigin=True))
	(S (VP*0 (VP*0 (PP (APPR 0) (ART 1) (NN 2)))) (VMFIN 3) (VP*1 (VP*1
		(CARD 4) (VVPP 5)) (VAINF 6))) """
	treeclass = tree.__class__
	for node in postorder(tree):
		nodes = list(node)
		node[:] = []
		for child in nodes:
			if disc(child):
				childnodes = list(child)
				child[:] = []
				node.extend(treeclass(('%s*%d' % (child.label, n)
						if markorigin else '%s*' % child.label), childsubset)
						for n, childsubset in enumerate(contsets(childnodes)))
			else:
				node.append(child)
	return canonicalize(tree)

SPLITLABEL = re.compile(r'(.*)\*(?:([0-9]+)([^!]+![^!]+)?)?$')


def mergediscnodes(tree):
	""" Reverse transformation of splitdiscnodes()

	>>> tree = Tree.parse('(S (VP (VP (PP (APPR 0) (ART 1) (NN 2)) (CARD 4)'
	... '(VVPP 5)) (VAINF 6)) (VMFIN 3))', parse_leaf=int)
	>>> print(mergediscnodes(splitdiscnodes(tree)))
	(S (VP (VP (PP (APPR 0) (ART 1) (NN 2)) (CARD 4) (VVPP 5)) (VAINF 6))
		(VMFIN 3))
	>>> print(mergediscnodes(splitdiscnodes(tree, markorigin=True)))
	(S (VP (VP (PP (APPR 0) (ART 1) (NN 2)) (CARD 4) (VVPP 5)) (VAINF 6))
		(VMFIN 3))
	>>> tree = Tree.parse('(S (X (A 0) (A 2)) (X (A 1) (A 3)))', parse_leaf=int)
	>>> print(mergediscnodes(splitdiscnodes(tree, markorigin=True)))
	(S (X (A 0) (A 2)) (X (A 1) (A 3)))
	>>> tree = Tree.parse('(S (X (A 0) (A 2)) (X (A 1) (A 3)))', parse_leaf=int)
	>>> print(splitdiscnodes(tree, markorigin=True))
	(S (X*0 (A 0)) (X*0 (A 1)) (X*1 (A 2)) (X*1 (A 3)))
	>>> tree = Tree.parse('(S (X (A 0) (A 2)) (X (A 1) (A 3)))', parse_leaf=int)
	>>> print(mergediscnodes(splitdiscnodes(tree)))
	(S (X (A 0) (A 1) (A 2) (A 3))) """
	treeclass = tree.__class__
	for node in tree.subtrees():
		merge = defaultdict(list)
		nodes = list(node)
		node[:] = []
		for child in nodes:
			if not isinstance(child, Tree):
				node.append(child)
				continue
			match = SPLITLABEL.search(child.label)
			if not match:
				node.append(child)
				continue
			grandchildren = list(child)
			child[:] = []
			if not merge[child.label]:
				merge[child.label].append(treeclass(match.group(1), []))
				node.append(merge[child.label][0])
			merge[child.label][0].extend(grandchildren)
			if match.group(2):
				nextlabel = '%s*%d' % (match.group(1), int(match.group(2)) + 1)
				merge[nextlabel].append(merge[child.label].pop(0))
	return tree


class OrderedSet(Set):
	""" A frozen, ordered set which maintains a regular list/tuple and set.
	The set is indexable. Equality is defined _without_ regard for order. """
	def __init__(self, iterable=None):
		if iterable:
			self.seq = tuple(iterable)
			self.theset = frozenset(self.seq)
		else:
			self.seq = ()
			self.theset = frozenset()

	def __hash__(self):
		return hash(self.theset)

	def __contains__(self, value):
		return value in self.theset

	def __len__(self):
		return len(self.theset)

	def __iter__(self):
		return iter(self.seq)

	def __getitem__(self, n):
		return self.seq[n]

	def __reversed__(self):
		return reversed(self.seq)

	def __repr__(self):
		if not self.seq:
			return '%s()' % self.__class__.__name__
		return '%s(%r)' % (self.__class__.__name__, self.seq)

	def __eq__(self, other):
		""" equality is defined _without_ regard for order. """
		return self.theset == set(other)

	def __and__(self, other):
		""" maintain the order of the left operand. """
		if not isinstance(other, Iterable):
			return NotImplemented
		return self._from_iterable(value for value in self if value in other)


def testminbin():
	""" Verify that all optimal parsing complexities are lower than or equal
	to the complexities of right-to-left binarizations. """
	from discodop.treebank import NegraCorpusReader
	corpus = NegraCorpusReader('.', 'alpinosample.export', punct='move')
	total = violations = violationshd = 0
	for n, tree, sent in zip(count(), list(
			corpus.parsed_sents().values())[:-2000], corpus.sents().values()):
		t = addbitsets(tree)
		if all(fanout(x) == 1 for x in t.subtrees()):
			continue
		print(n, tree, '\n', ' '.join(sent))
		total += 1
		optbin = optimalbinarize(tree.copy(True), headdriven=False, h=None, v=1)
		# undo head-ordering to get a normal right-to-left binarization
		normbin = addbitsets(binarize(canonicalize(Tree.convert(tree))))
		if (max(map(complexityfanout, optbin.subtrees()))
				> max(map(complexityfanout, normbin.subtrees()))):
			print('non-hd\n', tree)
			print(max(map(complexityfanout, optbin.subtrees())), optbin)
			print(max(map(complexityfanout, normbin.subtrees())), normbin, '\n')
			violations += 1

		optbin = optimalbinarize(tree.copy(True), headdriven=True, h=1, v=1)
		normbin = addbitsets(binarize(Tree.convert(tree), horzmarkov=1))
		if (max(map(complexityfanout, optbin.subtrees()))
				> max(map(complexityfanout, normbin.subtrees()))):
			print('hd\n', tree)
			print(max(map(complexityfanout, optbin.subtrees())), optbin)
			print(max(map(complexityfanout, normbin.subtrees())), normbin, '\n')
			violationshd += 1
	print('violations normal: %d / %d;  hd: %d / %d' % (
			violations, total, violationshd, total))
	assert violations == violationshd == 0


def testsplit():
	""" Verify that splitting and merging discontinuties gives the
	same trees for a treebank. """
	from discodop.treebank import NegraCorpusReader
	correct = wrong = 0
	n = NegraCorpusReader('.', 'alpinosample.export')
	for tree in n.parsed_sents().values():
		if mergediscnodes(splitdiscnodes(tree)) == tree:
			correct += 1
		else:
			wrong += 1
	total = len(n.sents())
	print("correct", correct, "=", 100 * correct / total, "%")
	print("wrong", wrong, "=", 100 * wrong / total, "%")


def test():
	""" Run all examples. """
	testminbin()
	testsplit()


def main():
	""" Command line interface for applying tree(bank) transforms. """
	import io
	from getopt import gnu_getopt, GetoptError
	from discodop.treebank import getreader, writetree, readheadrules, splitpath
	flags = ['markorigin', 'markheads', 'lemmas']
	options = ('inputfmt= outputfmt= inputenc= outputenc= slice= punct= '
			'headrules= functions= morphology= factor= markorigin=').split()
	try:
		opts, args = gnu_getopt(sys.argv[1:], 'h:v:', flags + options)
		assert 1 <= len(args) <= 3, 'expected 1, 2, or 3 positional arguments'
	except (GetoptError, AssertionError) as err:
		print('error: %r\n%s' % (err, USAGE))
		sys.exit(2)
	opts = dict(opts)
	action = args[0]
	infilename = args[1] if len(args) >= 2 else '/dev/stdin'
	outfilename = args[2] if len(args) == 3 else '/dev/stdout'

	# read input
	reader = getreader(opts.get('--inputfmt', 'export'))
	corpus = reader(*splitpath(infilename),
			encoding=opts.get('--inputenc', 'utf-8'),
			headrules=opts.get('--headrules'), markheads='--markheads' in opts,
			punct=opts.get('--punct'), functions=opts.get('--functions'),
			morphology=opts.get('--morphology'),
			lemmas='between' if '--lemmas' in opts else None)
	start, end = opts.get('--slice', ':').split(':')
	start = int(start) if start else None
	end = int(end) if end else None
	trees = list(corpus.parsed_sents().values())[start:end]
	sents = list(corpus.sents().values())[start:end]
	keys = list(corpus.sents())[start:end]
	print('read %d trees from %s' % (len(trees), infilename), file=sys.stderr)

	# apply transformation
	actions = ('binarize unbinarize optimalbinarize introducepreterminals '
			'splitdisc mergedisc none').split()
	assert action in actions, ('unrecognized action: %r\n'
			'available actions: %r' % (action, actions))
	if action in ('binarize', 'optimalbinarize'):
		factor = opts.get('--factor', 'right')
		headdriven = '--headrules' in opts
		h = int(opts['-h']) if 'h' in opts else None
		v = int(opts.get('-v', 1))
	for n, tree in enumerate(trees):
		if action == 'binarize':
			binarize(tree, factor, h, v)
		elif action == 'optimalbinarize':
			trees[n] = optimalbinarize(tree, '|', headdriven, h, v)
		elif action == 'unbinarize':
			unbinarize(tree)
		elif action == 'introducepreterminals':
			introducepreterminals(tree)
		elif action == 'splitdisc':
			splitdiscnodes(tree, '--markorigin' in opts)
		elif action == 'mergedisc':
			mergediscnodes(tree)
	if action != 'none':
		print('transformed %d trees with action %r' % (len(trees), action),
				file=sys.stderr)

	# write output
	headrules = None
	if opts.get('--outputfmt') in ('mst', 'conll'):
		assert opts.get('--headrules'), (
				'need head rules for dependency conversion')
		headrules = readheadrules(opts.get('--headrules'))

	print('going to write %d trees to %s' % (len(trees), outfilename),
			file=sys.stderr)
	encoding = opts.get('outputenc', 'utf-8')
	with io.open(outfilename, 'w', encoding=encoding) as outfile:
		if action == 'none' and (None == opts.get('--headrules')
				== opts.get('--markheads') == opts.get('--punct')
				and opts.get('--inputfmt') == opts.get('--outputfmt')):
			# copy treebank verbatim. useful when only taking a slice
			# or converting encoding.
			outfile.writelines(block for block in corpus.blocks().values())
		else:
			outfile.writelines(writetree(a, b, c,
					opts.get('--outputfmt', 'export'), headrules) for a, b, c
					in zip(trees, sents, keys))

if __name__ == '__main__':
	main()



# -*- coding: UTF-8 -*-
""" Treebank transformations:
- generic transforms listed by name
- Relational-realizational transform
- reattaching punctuation """
from __future__ import division, print_function, unicode_literals
from itertools import count, islice
from discodop.tree import Tree, ParentedTree

FIELDS = tuple(range(8))
WORD, LEMMA, TAG, MORPH, FUNC, PARENT, SECEDGETAG, SECEDGEPARENT = FIELDS
STATESPLIT = '^'


def transform(tree, sent, transformations):
	""" Perform some transformations, specific to Negra and WSJ treebanks.
	State-splits are preceded by '^'.
	negra:  transformations=('S-RC', 'VP-GF', 'NP', 'PUNCT')
	wsj:    transformations=('S-WH', 'VP-HD', 'S-INF')
	alpino: transformations=('PUNCT', ) """
	for name in transformations:
		# negra / tiger
		if name == 'S-RC':  # relative clause => S becomes SRC
			for s in tree.subtrees(lambda n: n.label == 'S'
					and function(n) == 'RC'):
				s.label += STATESPLIT + 'RC'
		elif name == 'NP':  # case
			for np in tree.subtrees(lambda n: n.label == 'NP'):
				np.label += STATESPLIT + function(np)
		elif name == 'PUNCT':  # distinguish sentence-ending punctuation.
			for punct in tree.subtrees(lambda n: isinstance(n[0], int)
					and sent[n[0]] in '.?!'):
				punct.label += STATESPLIT + sent[punct[0]]
		elif name == 'PP-NP':  # un-flatten PPs by introducing NPs
			addtopp = ('AC', )
			for pp in tree.subtrees(lambda n: n.label == 'PP'):
				ac = [a for a in pp if function(a) in addtopp]
				# anything before an initial preposition goes to the PP
				# (modifiers, punctuation), otherwise it goes to the NP;
				# mutatis mutandis for postpositions.
				functions = [function(x) for x in pp]
				if 'AC' in functions and 'NK' in functions:
					if functions.index('AC') < functions.index('NK'):
						ac[:0] = pp[:functions.index('AC')]
					if rindex(functions, 'AC') > rindex(functions, 'NK'):
						ac += pp[rindex(functions, 'AC') + 1:]
				#else:
				#	print('PP but no AC or NK', ' '.join(functions))
				nk = [a for a in pp if a not in ac]
				# introduce a PP unless there is already an NP in the PP
				# (annotation mistake?), or there is a PN and we want to avoid
				# a cylic unary of NP -> PN -> NP.
				if ac and nk and (len(nk) > 1
						or nk[0].label not in s('NP', 'PN')):
					pp[:] = []
					pp[:] = ac + [ParentedTree('NP', nk)]
		elif name == 'DP':  # introduce determiner phrases (DPs)
			#determiners = set('ART PDS PDAT PIS PIAT PPOSAT PRELS PRELAT '
			#	'PWS PWAT PWAV'.split())
			determiners = {'ART'}
			for np in list(tree.subtrees(lambda n: n.label == 'NP')):
				if np[0].label in determiners:
					np.label = 'DP'
					if len(np) > 1 and np[1].label != 'PN':
						np1 = np[1:]
						np[1:] = []
						np[1:] = [ParentedTree('NP', np1)]
		elif name == 'VP-GF':  # VP category split based on head
			for vp in tree.subtrees(lambda n: n.label == 'VP'):
				vp.label += STATESPLIT + function(vp)
		elif name == 'VP-FIN_NEGRA':  # introduce finite VP at S level
			# collect objects and modifiers
			# introduce new S level for discourse markers
			newlevel = 'DM'.split()
			addtovp = 'HD AC DA MO NG OA OA2 OC OG PD VO SVP'.split()

			def finitevp(s):
				""" Introduce finite VPs grouping verbs and their objects. """
				if any(x.label.startswith('V') and x.label.endswith('FIN')
						for x in s if isinstance(x, Tree)):
					vp = [a for a in s if function(a) in addtovp]
					# introduce a VP unless it would lead to a unary
					# VP -> VP production
					if len(vp) != 1 or vp[0].label != 'VP':
						s[:] = [pop(a) for a in s if function(a) not in addtovp
								] + [pop(a) for a in vp]
			toplevel_s = []
			if 'S' in labels(tree):
				toplevel_s = [a for a in tree if a.label == 'S']
				for s in toplevel_s:
					while function(s[0]) in newlevel:
						s[:] = [s[0], ParentedTree('S', s[1:])]
						s = s[1]
						toplevel_s = [s]
			elif 'CS' in labels(tree):
				cs = tree[labels(tree).index('CS')]
				toplevel_s = [a for a in cs if a.label == 'S']
			for a in toplevel_s:
				finitevp(a)
		elif name == 'POS-PART':  # introduce POS tag for particle verbs
			for a in tree.subtrees(
					lambda n: any(function(x) == 'SVP' for x in n)):
				svp = [x for x in a if function(x) == 'SVP'].pop()
				# apparently there can be a _verb_ particle without a verb.
				# headlines? annotation mistake?
				if any(map(ishead, a)):
					hd = [x for x in a if ishead(x)].pop()
					if hd.label != a.label:
						particleverb = ParentedTree(hd.label, [hd, svp])
						a[:] = [particleverb if ishead(x) else x
											for x in a if function(x) != 'SVP']
		elif name == 'SBAR':  # introduce SBAR level
			sbarfunc = 'CP'.split()
			# in the annotation, complementizers belong to the first S
			# in S conjunctions, even when they appear to apply to the whole
			# conjunction.
			for s in list(tree.subtrees(lambda n: n.label == 'S'
					and function(n[0]) in sbarfunc and len(n) > 1)):
				s.label = 'SBAR'
				s[:] = [s[0], ParentedTree('S', s[1:])]
		elif name == 'NEST':  # introduce nested structures for modifiers
			# (iterated adjunction instead of sister adjunction)
			adjunctable = set('NP'.split())  # PP AP VP
			for a in list(tree.subtrees(lambda n: n.label in adjunctable
					and any(function(x) == 'MO' for x in n))):
				modifiers = [x for x in a if function(x) == 'MO']
				if min(n for n, x in enumerate(a) if function(x) == 'MO') == 0:
					modifiers[:] = modifiers[::-1]
				while modifiers:
					modifier = modifiers.pop()
					a[:] = [ParentedTree(a.label,
							[x for x in a if x != modifier]), modifier]
					a = a[0]
		elif name == 'UNARY':  # introduce phrasal projections for single tokens
			# currently only adds NPs.
			tagtoconst, _ = getgeneralizations()
			for a in tree.treepositions('leaves'):
				tag = tree[a[:-1]]   # e.g. NN
				const = tree[a[:-2]]  # e.g. S
				if (tag.label in tagtoconst
						and const.label != tagtoconst[tag.label]):
					# NN -> NN -> word
					tag[:] = [ParentedTree(tag.label, [tag[0]])]
					# NP -> NN -> word
					tag.source = 8 * ['--']
					tag.source[TAG] = tag.label = tagtoconst[tag.label]
					tag.source[FUNC] = function(tag[0])
					#tag[0].source[FUNC] = 'NK'

		# wsj
		elif name == 'S-WH':
			for sbar in tree.subtrees(lambda n: n.label == 'SBAR'):
				for s in sbar:
					if (s.label == 'S'
							and any(a.label.startswith('WH') for a in s)):
						s.label += STATESPLIT + 'WH'
		elif name == 'VP-HD':  # VP category split based on head
			for vp in tree.subtrees(lambda n: n.label == 'VP'):
				hd = [x for x in vp if ishead(x)].pop()
				if hd.label == 'VB':
					vp.label += STATESPLIT + 'HINF'
				elif hd.label == 'TO':
					vp.label += STATESPLIT + 'HTO'
				elif hd.label in ('VBN', 'VBG'):
					vp.label += STATESPLIT + 'HPART'
		elif name == 'S-INF':
			for s in tree.subtrees(lambda n: n.label == 'S'):
				hd = [x for x in s if ishead(x)].pop()
				if hd.label in ('VP' + STATESPLIT + 'HINF',
						'VP' + STATESPLIT + 'HTO'):
					s.label += STATESPLIT + 'INF'
		elif name == 'VP-FIN_WSJ':  # add disc. finite VP when verb is under S
			for s in tree.subtrees(lambda n: n.label == 'S'):
				if not any(a.label.startswith('VP') for a in s):
					raise NotImplementedError
		# alpino?
		# ...
		else:
			raise ValueError('unrecognized transformation %r' % name)
	for a in reversed(list(tree.subtrees(lambda x: len(x) > 1))):
		a.sort(key=Tree.leaves)
	return tree


def reversetransform(tree, transformations):
	""" Undo specified transformations, as well as removing any hyphen-marked
	state splits. Do not apply twice (might remove VPs which shouldn't be). """
	tagtoconst, _ = getgeneralizations()
	# Generic state-split removal
	for node in tree.subtrees(lambda n: STATESPLIT in n.label[1:]):
		node.label = node.label[:node.label.index(STATESPLIT, 1)]

	# restore linear precedence ordering
	for a in tree.subtrees(lambda n: len(n) > 1):
		a.sort(key=lambda n: n.leaves())
	for name in reversed(transformations):
		if name == 'DP':  # remove DPs
			for dp in tree.subtrees(lambda n: n.label == 'DP'):
				dp.label = 'NP'
				if len(dp) > 1 and dp[1].label == 'NP':
					#dp1 = dp[1][:]
					#dp[1][:] = []
					#dp[1:] = dp1
					dp[1][:], dp[1:] = [], dp[1][:]
		elif name == 'NEST':  # flatten adjunctions
			nkonly = set('PDAT CAP PPOSS PPOSAT ADJA FM PRF NM NN NE PIAT '
					'PRELS PN TRUNC CH CNP PWAT PDS VP CS CARD ART PWS PPER'
					''.split())
			probably_nk = set('AP PIS'.split()) | nkonly
			for np in tree.subtrees(lambda n: len(n) == 2
					and n.label == 'NP'
					and [x.label for x in n].count('NP') == 1
					and not set(labels(n)) & probably_nk):
				np.sort(key=lambda n: n.label == 'NP')
				np[:] = np[:1] + np[1][:]
		elif name == 'PP-NP':  # flatten PPs
			for pp in tree.subtrees(lambda n: n.label == 'PP'):
				if 'NP' in labels(pp) and 'NN' not in labels(pp):
					# ensure NP is in last position
					pp.sort(key=lambda n: n.label == 'NP')
					pp[-1][:], pp[-1:] = [], pp[-1][:]
		elif name == 'SBAR':  # merge extra S level
			for sbar in list(tree.subtrees(lambda n: n.label == 'SBAR'
					or (n.label == 'S' and len(n) == 2
						and labels(n) == ['PTKANT', 'S']))):
				sbar.label = 'S'
				if sbar[0].label == 'S':
					sbar[:] = sbar[1:] + sbar[0][:]
				else:
					sbar[:] = sbar[:1] + sbar[1][:]
		elif name == 'VP-FIN_NEGRA':
			def mergevp(s):
				""" merge finite VP with S level """
				for vp in (n for n, a in enumerate(s) if a.label == 'VP'):
					if any(a.label.endswith('FIN') for a in s[vp]):
						s[vp][:], s[vp:vp + 1] = [], s[vp][:]
			#if any(a.label == 'S' for a in tree):
			#	map(mergevp, [a for a in tree if a.label == 'S'])
			#elif any(a.label == 'CS' for a in tree):
			#	map(mergevp, [s for cs in tree for s in cs if cs.label == 'CS'
			#		and s.label == 'S'])
			for s in tree.subtrees(lambda n: n.label == 'S'):
				mergevp(s)
		elif name == 'POS-PART':
			# remove constituents for particle verbs
			# get the grandfather of each verb particle
			hasparticle = lambda n: any('PTKVZ' in (x.label
					for x in m if isinstance(x, Tree)) for m in n
					if isinstance(m, Tree))
			for a in list(tree.subtrees(hasparticle)):
				for n, b in enumerate(a):
					if (len(b) == 2 and b.label.startswith('V')
						and 'PTKVZ' in (c.label for c in b
							if isinstance(c, Tree))
						and any(c.label == b.label for c in b)):
						a[n:n + 1] = b[:]
		elif name == 'UNARY':
			# remove phrasal projections for single tokens
			for a in tree.treepositions('leaves'):
				tag = tree[a[:-1]]    # NN
				const = tree[a[:-2]]  # NP
				parent = tree[a[:-3]]  # PP
				if (len(const) == 1 and tag.label in tagtoconst
						and const.label == tagtoconst[tag.label]):
					parent[a[-3]] = const.pop(0)
					del const
	# restore linear precedence ordering
	for a in tree.subtrees(lambda n: len(n) > 1):
		a.sort(key=lambda n: n.leaves())
	return tree


def getgeneralizations():
	""" Some tree transforms for Negra/Tiger.  """
	# generalizations suggested by SyntaxGeneralizer of TigerAPI
	# however, instead of renaming, we introduce unary productions
	# POS tags
	tonp = 'NN NNE PNC PRF PDS PIS PPER PPOS PRELS PWS'.split()
	#topp = 'PROAV PWAV'.split()  # ??
	#toap = 'ADJA PDAT PIAT PPOSAT PRELAT PWAT PRELS'.split()
	#toavp = 'ADJD ADV'.split()

	tagtoconst = {}
	for label in tonp:
		tagtoconst[label] = 'NP'
	#for label in toap:
	#	tagtoconst[label] = 'AP'
	#for label in toavp:
	#	tagtoconst[label] = 'AVP'

	# phrasal categories
	tonp = 'CNP NM PN'.split()
	#topp = 'CPP'.split()
	#toap = 'MTA CAP'.split()
	#toavp = 'AA CAVP'.split()
	unaryconst = {}
	for label in tonp:
		unaryconst[label] = 'NP'
	return tagtoconst, unaryconst


def unifymorphfeat(feats, percolatefeatures=None):
	""" Treat a sequence of strings as feature vectors, either
	comma or dot separated, and produce the sorted union of their features.

	:param percolatefeatures: if a set is given, select only these features;
		by default all features are used.

	>>> print(unifymorphfeat({'Def.*.*', '*.Sg.*', '*.*.Akk'}))
	Akk.Def.Sg
	>>> print(unifymorphfeat({'LID[bep,stan,rest]', 'N[soort,ev,zijd,stan]'}))
	bep,ev,rest,soort,stan,zijd """
	sep = '.' if any('.' in a for a in feats) else ','
	result = set()
	for a in feats:
		if '[' in a:
			a = a[a.index('[') + 1:a.index(']')]
		result.update(a.split(sep))
	if percolatefeatures:
		result.intersection_update(percolatefeatures)
	return sep.join(sorted(result - {'*', '--'}))


def rrtransform(tree, morphlevels=0, percolatefeatures=None,
		adjunctionlabel=None, ignorefunctions=None, ignorecategories=None):
	""" Relational-realizational tree transformation.
	Every constituent node is expanded to three levels:

	1) syntactic category, e.g., S
	2) unordered functional argument structure of children, e.g., S/<SBJ,HD,OBJ>
	3) for each child:
		grammatical function + parent syntactic category, e.g., OBJ/S

	Example::

		(NP-SBJ (NN-HD ...)) => (NP (<HD>/NP (HD/NP (NN ...))))

	:param adjunctionlabel: a grammatical function label identifying
		adjunctions. They will not be part of argument structures, and their
		grammatical function will be replaced with their neighboring
		non-adjunctive functions.
	:param ignorefunctions: function labels that do not go into argument
		structure, but keep their function in their realization to make
		backtransform possible.
	:param morphlevels: if nonzero, percolate morphological features this many
		levels upwards. For a given node, the union of the features of its
		children are collected, and the result is appended to its syntactic
		category.
	:param percolatefeatures: if a sequence is given, percolate only these
		morphological features; by default all features are used.
	:returns: a new, transformed tree. """
	def realize(child, prevfunc, nextfunc):
		""" Generate realization of a child node by recursion. """
		newchild, morph, lvl = rrtransform(child, morphlevels,
				percolatefeatures, adjunctionlabel, ignorefunctions,
				ignorecategories)
		result = tree.__class__('%s/%s' % (('%s:%s' % (prevfunc, nextfunc)
				if child.source[FUNC] == adjunctionlabel
				else child.source[FUNC]), tree.label), [newchild])
		return result, morph, lvl

	if not isinstance(tree[0], Tree):
		morph = tree.source[MORPH].replace('(', '[').replace(')', ']')
		preterminal = tree.__class__('%s/%s' % (tree.label, morph), tree)
		if morphlevels:
			return preterminal, morph, morphlevels
		return preterminal, None, 0
	# for each node, collect the functions of closest non-adjunctive sibling
	childfuncsl = (prevfunc, ) = ['']
	for child in tree:
		if (isinstance(child, Tree) and child.source[FUNC]
				and child.source[FUNC] != adjunctionlabel
				and child.source[FUNC] not in ignorefunctions
				and child.label not in ignorecategories):
			prevfunc = child.source[FUNC]
		childfuncsl.append(prevfunc)
	childfuncsr = (nextfunc, ) = ['']
	for child in reversed(tree[1:]):
		if (isinstance(child, Tree) and child.source[FUNC]
				and child.source[FUNC] != adjunctionlabel
				and child.source[FUNC] not in ignorefunctions
				and child.label not in ignorecategories):
			nextfunc = child.source[FUNC]
		childfuncsr.insert(0, prevfunc)
	funcstr = ','.join(sorted(child.source[FUNC] for child in tree
			if isinstance(child, Tree) and child.source[FUNC]
					and child.source[FUNC] != adjunctionlabel
					and child.source[FUNC] not in ignorefunctions
					and child.label not in ignorecategories))
	children, feats, levels = [], [], []
	for child, prevfunc, nextfunc in zip(tree, childfuncsl, childfuncsr):
		newchild, morph, lvl = realize(child, prevfunc, nextfunc)
		children.append(newchild)
		if morph and lvl:
			feats.append(morph)
			levels.append(lvl)
	morph, lvl = None, 0
	if feats and max(levels) and tree.label != 'ROOT':
		morph, lvl = unifymorphfeat(feats, percolatefeatures), max(levels) - 1
	configuration = tree.__class__('%s/<%s>' % (tree.label, funcstr),
			children)
	projection = tree.__class__(('%s-%s' % (tree.label, morph)) if morph
			else tree.label, [configuration])
	return projection, morph, lvl


def rrbacktransform(tree, adjunctionlabel=None, func=None):
	""" Reverse the relational-realizational transformation, conserving
	grammatical functions.

	:param adjunctionlabel: used to assign a grammatical function to
		adjunctions that have been converted to contextual labels 'next:prev'.
	:param func: used internally to percolate functional labels.
	:returns: a new tree. """
	morph = None
	if not isinstance(tree[0], Tree):
		tag, morph = tree.label.split('/')
		result = tree.__class__(tag, tree)
	elif '/' not in tree[0].label:
		result = tree.__class__(tree.label,
				[rrbacktransform(child, adjunctionlabel) for child in tree])
	else:
		result = tree.__class__(tree.label.split('-')[0],
				[rrbacktransform(
						child[0],
						adjunctionlabel,
						child.label.split('/')[0])
					for child in tree[0]])
	result.source = ['--'] * 8
	result.source[TAG] = result.label
	if morph:
		result.source[MORPH] = morph.replace('[', '(').replace(']', ')')
	if func and adjunctionlabel and ':' in func:
		result.source[FUNC] = adjunctionlabel
	elif func:
		result.source[FUNC] = func
	return result


# relative frequencies of punctuation in Negra: (is XY an annotation error?)
#1/1             $,      ,
#14793/17269     $.      .
#8598/13345      $[      "
#1557/13345      $[      )
#1557/13345      $[      (
#1843/17269      $.      :
#232/2669        $[      -
#343/13345       $[      /
#276/17269       $.      ?
#249/17269       $.      ;
#89/13345        $[      '
#101/17269       $.      !
#2/513           XY      :
#41/13345        $[      ...
#1/513           XY      -
#1/2467          $.      ·      # NB this is not a period but a \cdot ...


PUNCTUATION = frozenset('.,():\'-";?/!*&`[]<>{}|=\'\xc2\xab\xc2\xbb\xb7\xad\\'
		) | {'&bullet;', '..', '...', '....', '.....', '......', '!!', '!!!',
		'??', '???', "''", '``', u',,', '--'}


def ispunct(word, tag):
	""" Test whether a word and/or tag is punctuation. """
	# fixme: treebank specific parameters for detecting punctuation.
	return (tag in ('$,', '$.', '$[', '$(',
			'let', 'LET[]', 'SPEC[symb]', 'TW[hoofd,vrij]')
			or word in PUNCTUATION)


def punctremove(tree, sent):
	""" Remove any punctuation nodes, and any empty ancestors. """
	for a in reversed(tree.treepositions('leaves')):
		if ispunct(sent[tree[a]], tree[a[:-1]].label):
			# remove this punctuation node and any empty ancestors
			for n in range(1, len(a)):
				del tree[a[:-n]]
				if tree[a[:-(n + 1)]]:
					break
	# renumber
	oldleaves = sorted(tree.leaves())
	newleaves = {a: n for n, a in enumerate(oldleaves)}
	for a in tree.treepositions('leaves'):
		tree[a] = newleaves[tree[a]]
	assert sorted(tree.leaves()) == list(range(len(tree.leaves()))), tree


def punctroot(tree, sent):
	""" Move punctuation directly under the ROOT node, as in the
	original Negra/Tiger treebanks. """
	punct = []
	for a in reversed(tree.treepositions('leaves')):
		if ispunct(sent[tree[a]], tree[a[:-1]].label):
			# store punctuation node
			punct.append(tree[a[:-1]])
			# remove this punctuation node and any empty ancestors
			for n in range(1, len(a)):
				del tree[a[:-n]]
				if tree[a[:-(n + 1)]]:
					break
	tree.extend(punct)


def punctlower(tree, sent):
	""" Find suitable constituent for punctuation marks and add it there;
	removal at previous location is up to the caller.  Based on rparse code.
	Initial candidate is the root node. """
	def lower(node, candidate):
		""" Lower a specific instance of punctuation in tree,
		recursing top-down on suitable candidates. """
		num = node.leaves()[0]
		for i, child in enumerate(sorted(candidate, key=lambda x: x.leaves())):
			if not isinstance(child[0], Tree):
				continue
			termdom = child.leaves()
			if num < min(termdom):
				print('moving', node, 'under', candidate.label)
				candidate.insert(i + 1, node)
				break
			elif num < max(termdom):
				lower(node, child)
				break

	for a in tree.treepositions('leaves'):
		if ispunct(sent[tree[a]], tree[a[:-1]].label):
			b = tree[a[:-1]]
			del tree[a[:-1]]
			lower(b, tree)


def punctraise(tree, sent):
	""" Trees in the Negra corpus have punctuation attached to the root;
	i.e., it is not part of the phrase-structure.  This function attaches
	punctuation nodes (that is, a POS tag with punctuation terminal) to an
	appropriate constituent. """
	#punct = [node for node in tree.subtrees() if isinstance(node[0], int)
	punct = [node for node in tree if isinstance(node[0], int)
			and ispunct(sent[node[0]], node.label)]
	while punct:
		node = punct.pop()
		# dedicated punctation node (??)
		if all(isinstance(a[0], int) and ispunct(sent[a[0]], a)
				for a in node.parent):
			continue
		node.parent.pop(node.parent_index)
		phrasalnode = lambda x: len(x) and isinstance(x[0], Tree)
		for candidate in tree.subtrees(phrasalnode):
			# add punctuation mark next to biggest constituent which it borders
			#if any(node[0] - 1 in borders(sorted(a.leaves())) for a in candidate):
			#if any(node[0] - 1 == max(a.leaves()) for a in candidate):
			if any(node[0] + 1 == min(a.leaves()) for a in candidate):
				candidate.append(node)
				break
		else:
			tree.append(node)

BALANCEDPUNCTMATCH = {'"': '"', '[': ']', '(': ')', '-': '-', "'": "'",
		'\xc2\xab': '\xc2\xbb'}  # the last ones are unicode for << and >>.


def balancedpunctraise(tree, sent):
	""" Move balanced punctuation marks " ' - ( ) [ ] together in the same
	constituent. Based on rparse code.

	>>> tree = ParentedTree.parse('(ROOT ($, 3) ($[ 7) ($[ 13) ($, 14) ($, 20)'
	... ' (S (NP (ART 0) (ADJA 1) (NN 2) (NP (CARD 4) (NN 5) (PP (APPR 6) '
	... '(CNP (NN 8) (ADV 9) (ISU ($. 10) ($. 11) ($. 12))))) (S (PRELS 15) '
	... '(MPN (NE 16) (NE 17)) (ADJD 18) (VVFIN 19))) (VVFIN 21) (ADV 22) '
	... '(NP (ADJA 23) (NN 24))) ($. 25))', parse_leaf=int)
	>>> sent = ("Die zweite Konzertreihe , sechs Abende mit ' Orgel plus "
	... ". . . ' , die Hayko Siemens musikalisch leitet , bietet wieder "
	... "ungewoehnliche Kombinationen .".split())
	>>> punctraise(tree, sent)
	>>> balancedpunctraise(tree, sent)
	>>> from discodop.treetransforms import addbitsets, fanout
	>>> max(map(fanout, addbitsets(tree).subtrees()))
	1
	>>> nopunct = Tree.parse('(ROOT (S (NP (ART 0) (ADJA 1) (NN 2) (NP '
	... '(CARD 3) (NN 4) (PP (APPR 5) (CNP (NN 6) (ADV 7)))) (S (PRELS 8) '
	... '(MPN (NE 9) (NE 10)) (ADJD 11) (VVFIN 12))) (VVFIN 13) (ADV 14) '
	... '(NP (ADJA 15) (NN 16))))', parse_leaf=int)
	>>> max(map(fanout, addbitsets(nopunct).subtrees()))
	1
	"""
	assert isinstance(tree, ParentedTree)
	# right punct str as key, mapped to left index as value
	punctmap = {}
	# punctuation indices mapped to preterminal nodes
	termparent = {a[0]: a for a in tree.subtrees()
			if a and isinstance(a[0], int) and ispunct(sent[a[0]], a.label)}
	for terminal in sorted(termparent):
		preterminal = termparent[terminal]
		# do we know the matching punctuation mark for this one?
		if sent[terminal] in punctmap:
			right = terminal
			left = punctmap[sent[right]]
			rightparent = preterminal.parent
			leftparent = termparent[left].parent
			if max(leftparent.leaves()) == right - 1:
				node = termparent[right]
				leftparent.append(node.parent.pop(node.parent_index))
			elif min(rightparent.leaves()) == left + 1:
				node = termparent[left]
				rightparent.insert(0, node.parent.pop(node.parent_index))
			if sent[right] in punctmap:
				del punctmap[sent[right]]
		elif sent[terminal] in BALANCEDPUNCTMATCH:
			punctmap[BALANCEDPUNCTMATCH[sent[terminal]]] = terminal


def function(tree):
	""" :returns: grammatical function for node, or an empty string. """
	if getattr(tree, 'source', None):
		return tree.source[FUNC].split('-')[0]
	return ''


def ishead(tree):
	""" Test whether this node is the head of the parent constituent. """
	if getattr(tree, 'source', None):
		return 'HD' in tree.source[FUNC].upper().split('-')
	return False


def rindex(l, v):
	""" Like list.index(), but go from right to left. """
	return len(l) - 1 - l[::-1].index(v)


def labels(tree):
	""" :returns: the labels of the children of this node. """
	return [a.label for a in tree if isinstance(a, Tree)]


def pop(a):
	""" Remove this node from its parent node, if it has one.
	Convenience function for ParentedTrees. """
	try:
		return a.parent.pop(a.parent_index)
	except AttributeError:
		return a


def bracketings(tree):
	""" Labelled bracketings of a tree. """
	return [(a.label, tuple(sorted(a.leaves())))
		for a in tree.subtrees(lambda t: t and isinstance(t[0], Tree))]


def testpunct():
	""" Verify that punctuation movement does not increase fan-out. """
	from discodop.treetransforms import addbitsets, fanout
	from discodop.treebank import NegraCorpusReader
	filename = 'alpinosample.export'
	mangledtrees = NegraCorpusReader('.', filename, punct='move')
	nopunct = list(NegraCorpusReader('.', filename,
			punct='remove').parsed_sents().values())
	originals = list(NegraCorpusReader('.', filename, headrules=None,
			encoding='iso-8859-1').parsed_sents().values())
	phrasal = lambda x: len(x) and isinstance(x[0], Tree)
	for n, mangled, sent, nopunct, original in zip(count(),
			mangledtrees.parsed_sents().values(),
			mangledtrees.sents().values(), nopunct, originals):
		print(n, end='')
		for a, b in zip(sorted(addbitsets(mangled).subtrees(phrasal),
				key=lambda n: min(n.leaves())),
				sorted(addbitsets(nopunct).subtrees(phrasal),
				key=lambda n: min(n.leaves()))):
			if fanout(a) != fanout(b):
				print(' '.join(sent))
				print(mangled)
				print(nopunct)
				print(original)
			assert fanout(a) == fanout(b), '%d %d\n%s\n%s' % (
				fanout(a), fanout(b), a, b)
	print()


def testtransforms():
	""" Test whether the Tiger transformations (transform / reversetransform)
	are reversible. """
	from discodop.treetransforms import canonicalize
	from discodop.treebank import NegraCorpusReader, handlefunctions
	headrules = None  # 'alpino.headrules'
	n = NegraCorpusReader('.', 'alpinosample.export', headrules=headrules)
	nn = NegraCorpusReader('.', 'alpinosample.export', headrules=headrules)
	transformations = ('S-RC', 'VP-GF', 'NP')
	trees = [transform(tree, sent, transformations)
			for tree, sent in zip(nn.parsed_sents().values(),
				nn.sents().values())]
	print('\ntransformed')
	correct = exact = d = 0
	for a, b, c in islice(zip(n.parsed_sents().values(),
			trees, n.sents().values()), 100):
		transformb = reversetransform(b.copy(True), transformations)
		b1 = bracketings(canonicalize(a))
		b2 = bracketings(canonicalize(transformb))
		z = -1  # 825
		if b1 != b2 or d == z:
			precision = len(set(b1) & set(b2)) / len(set(b1))
			recall = len(set(b1) & set(b2)) / len(set(b2))
			if precision != 1.0 or recall != 1.0 or d == z:
				print(d, ' '.join(':'.join((str(n),
					a.encode('unicode-escape'))) for n, a in enumerate(c)))
				print('no match', precision, recall)
				print(len(b1), len(b2), 'gold-transformed', set(b2) - set(b1),
						'transformed-gold', set(b1) - set(b2))
				print(a)
				print(transformb)
				handlefunctions('add', a)
				print(a)
				print(b)
				print()
			else:
				correct += 1
		else:
			exact += 1
			correct += 1
		d += 1
	print('matches', correct, '/', d, 100 * correct / d, '%')
	print('exact', exact)

if __name__ == '__main__':
	main()


def main():
	""" Just tests. """
	testpunct()
	testtransforms()

if __name__ == '__main__':
	main()



""" Web interface to browse through a treebank. Requires Flask. """
# stdlib
import os
import logging
#from functools import wraps
# Flask & co
from flask import Flask
from flask import request, render_template
from flask import send_from_directory
#from werkzeug.contrib.cache import SimpleCache
# disco-dop
#from discodop.treebank import getreader
from discodop.treedraw import DrawTree

APP = Flask(__name__)


def stream_template(template_name, **context):
	""" From Flask documentation. """
	APP.update_template_context(context)
	templ = APP.jinja_env.get_template(template_name)
	result = templ.stream(context)
	result.enable_buffering(5)
	return result


@APP.route('/')
def main():
	""" Entry form to upload file. """
	return render_template('upload.html', form=request.args)


@APP.route('/view')
def view():
	""" Receive uploaded file, store it with timeout, and produce chunk of 100
	trees to display. """
	pass


@APP.route('/draw')
def draw():
	""" Produce a visualization of a tree on a separate page. """
	if 'tree' in request.args:
		return "<pre>%s</pre>" % DrawTree(request.args['tree']).text(
				unicodelines=True, html=True)


@APP.route('/favicon.ico')
def favicon():
	""" Serve the favicon. """
	return send_from_directory(os.path.join(APP.root_path, 'static'),
			'treesearch.ico', mimetype='image/vnd.microsoft.icon')


if __name__ == '__main__':
	logging.basicConfig()
	for log in (logging.getLogger(), APP.logger):
		log.setLevel(logging.DEBUG)
		log.handlers[0].setFormatter(logging.Formatter(
				fmt='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S'))
	APP.run(debug=True, host='0.0.0.0')
